from __future__ import print_function, division
import numpy as np

from astropy.io import fits
from astropy.time import Time
import datetime, time

import logging
_log = logging.getLogger('pynrc')

from . import conf
from .logging_utils import setup_logging


from webbpsf_ext.webbpsf_ext_core import _check_list
from .nrc_utils import pix_noise



class multiaccum(object):
    """
    A class for defining MULTIACCUM ramp settings.
    See `NIRCam MULTIACCUM documentation
    <https://jwst-docs.stsci.edu/display/JTI/NIRCam+Detector+Readout+Patterns>`_
    for more details.

    Parameters
    ----------------
    read_mode : str
        NIRCam Ramp Readout mode such as 'RAPID', 'BRIGHT1', 'DEEP8', etc., or 'CUSTOM'
    nint : int
        Number of integrations (ramps).
    ngroup : int
        Number of groups in a integration.
    nf : int
        Number of frames per group.
    nd1 : int
        Number of drop frame after reset (before first group read). Default=0.
    nd2 : int
        Number of drop frames within a group (ie., groupgap). 
    nd3 : int
        Number of drop frames after final read frame in ramp. Default=1.
    nr1 : int
        Number of reset frames within first ramp. Default=0.
    nr2 : int
        Number of reset frames for subsequent ramps. Default=1.
    wind_mode : str
        Set to determine maximum number of allowed groups.

    Notes
    -----

    **NIRCam-specific readout modes**
    
    ========  ===  ===
    Pattern    NF  ND2
    ========  ===  ===
    RAPID      1    0
    BRIGHT1    1    1
    BRIGHT2    2    0
    SHALLOW2   2    3
    SHALLOW4   4    1
    MEDIUM2    2    8
    MEDIUM8    8    2
    DEEP2      2   18
    DEEP8      8   12
    ========  ===  ===
    """

    def __init__(self, read_mode='RAPID', nint=1, ngroup=1, nf=1, nd1=0, nd2=0, nd3=0, 
                 nr1=1, nr2=1, wind_mode='FULL', **kwargs):

        self.nint = nint
        self._ngroup_max = 10000
        self.ngroup = ngroup

        # Modify these directly rather via the @property
        self._nr1 = self._check_int('nr1',nr1,0)
        self._nr2 = self._check_int('nr2',nr2,0)
        self._nf = self._check_int('nf',nf,1)
        self._nd1 = self._check_int('nd1',nd1,0)
        self._nd2 = self._check_int('nd2',nd2,0)
        self._nd3 = self._check_int('nd3',nd3,0)
        # Now set read mode to specified mode, which may modify nf, nd1, nd2, and nd3
        self.read_mode = read_mode

    @property
    def nint(self):
        """Number of ramps (integrations) in an exposure."""
        return self._nint
    @nint.setter
    def nint(self, value):
        self._nint = self._check_int('nint',value,1)

    @property
    def ngroup(self):
        """Number of groups in a ramp (integration)."""
        return self._ngroup
    @ngroup.setter
    def ngroup(self, value):
        value = self._check_int('ngroup',value,1)
        if value > self._ngroup_max:
            _log.warning('Specified ngroup ({}) greater than allowed value ({})'\
                         .format(value, self._ngroup_max))
            _log.warning('Setting ngroup = {}'.format(self._ngroup_max))
            value = self._ngroup_max
        self._ngroup = value

    @property
    def nf(self):
        """Number of frames per group."""
        return self._nf
    @nf.setter
    def nf(self, value):
        value = self._check_int('nf',value,1)
        self._nf = self._check_custom(value, self._nf)

    @property
    def nd1(self):
        """Number of drop frame after reset (before first group read)."""
        return self._nd1
    @nd1.setter
    def nd1(self, value):
        value = self._check_int('nd1',value,0)
        self._nd1 = self._check_custom(value, self._nd1)

    @property
    def nd2(self):
        """Number of drop frames within a group (aka, groupgap)."""
        return self._nd2
    @nd2.setter
    def nd2(self, value):
        value = self._check_int('nd2',value,0)
        self._nd2 = self._check_custom(value, self._nd2)

    @property
    def nd3(self):
        """Number of drop frames after final read frame in ramp."""
        return self._nd3
    @nd3.setter
    def nd3(self, value):
        value = self._check_int('nd3',value,0)
        self._nd3 = self._check_custom(value, self._nd3)

    @property
    def nr1(self):
        """Number of reset frames before first integration."""
        return self._nr1
    @nr1.setter
    def nr1(self, value):
        self._nr1 = self._check_int('nr1', value, minval=0)

    @property
    def nr2(self):
        """Number of reset frames for subsequent integrations."""
        return self._nr2
    @nr2.setter
    def nr2(self, value):
        self._nr2 = self._check_int('nr2', value, minval=0)

    @property
    def nread_tot(self):
        """Total number of read frames in a ramp, including drops"""
        nf = self.nf; nd1 = self.nd1; nd2 = self.nd2; nd3 = self.nd3
        ngroup = self.ngroup

        nframes = nd1 + ngroup*nf + (ngroup-1)*nd2 + nd3
        return nframes

    @property
    def read_mode(self):
        """Selected Read Mode in the `patterns_list` attribute."""
        return self._read_mode
    @read_mode.setter
    def read_mode(self, value):
        """Set MULTIACCUM Readout. Automatically updates other relevant attributes."""
        if value is None:
            _log.info('Readout pattern has None value. Setting to CUSTOM.')
            value = 'CUSTOM'

        value = value.upper()
        _check_list(value, self.patterns_list, var_name='read_mode')

        self._read_mode = value
        self._validate_readout()

    @property
    def patterns_list(self):
        """Allowed NIRCam MULTIACCUM patterns"""
        plist = sorted(list(self._pattern_settings.keys()))
        return ['CUSTOM'] + plist

    @property
    def _pattern_settings(self):
        patterns = ['RAPID', 'BRIGHT1', 'BRIGHT2', 'SHALLOW2', 'SHALLOW4', 'MEDIUM2', 'MEDIUM8', 'DEEP2', 'DEEP8']
        nf_arr   = [1,1,2,2,4,2,8, 2, 8]
        nd2_arr  = [0,1,0,3,1,8,2,18,12]
        ng_max   = [10,10,10,10,10,10,10,20,20]

        return dict(zip(patterns, zip(nf_arr, nd2_arr, ng_max)))

    def to_dict(self, verbose=False):
        """Export ramp settings to a dictionary."""
        p = [('read_mode',self.read_mode), ('nint',self.nint), ('ngroup',self.ngroup), \
             ('nf',self.nf), ('nd1',self.nd1), ('nd2',self.nd2), ('nd3',self.nd3)]
        return tuples_to_dict(p, verbose)


    def _validate_readout(self):
        """ 
        Validation to make sure the defined ngroups, nf, etc. are consistent with
        the selected MULTIACCUM readout pattern.
        """

        if self.read_mode not in self.patterns_list:
            _log.warning('Readout {} not a valid NIRCam readout mode. Setting to CUSTOM.'\
                         .format(self.read_mode))
            self._read_mode = 'CUSTOM'
            _log.warning('Using explicit settings: ngroup={}, nf={}, nd1={}, nd2={}, nd3={}'\
                         .format(self.ngroup, self.nf, self.nd1, self.nd2, self.nd3))
        elif self.read_mode == 'CUSTOM':
            _log.info('{} readout mode selected.'.format(self.read_mode))
            _log.info('Using explicit settings: ngroup={}, nf={}, nd1={}, nd2={}, nd3={}'\
                      .format(self.ngroup, self.nf, self.nd1, self.nd2, self.nd3))
        else:
            _log.info('{} readout mode selected.'.format(self.read_mode))
            nf, nd2, _ = self._pattern_settings.get(self.read_mode)
            self._nf  = nf
            self._nd1 = 0
            self._nd2 = nd2
            self._nd3 = 0
            _log.info('Setting ngroup={}, nf={}, nd1={}, nd2={}, nd3={}.'\
                     .format(self.ngroup, self.nf, self.nd1, self.nd2, self.nd3))


    def _check_custom(self, val_new, val_orig):
        """Check if read_mode='CUSTOM' before changing variable."""
        if self.read_mode == 'CUSTOM': 
            return val_new
        else: 
            _log.warning("Can only modify parameter if read_mode='CUSTOM'.")
            return val_orig

    def _check_int(self, pstr, val, minval=1):
        """Check if a value is a positive integer, otherwise throw exception."""
        val = float(val)
        if (val.is_integer()) and (val>=minval): 
            return int(val)
        else:
            raise ValueError("{}={} must be an integer >={}.".format(pstr,val,minval))


    
class det_timing(object):
    """ 
    Class to hold detector operations information. Includes SCA attributes such as
    detector names and IDs as well as :class:`multiaccum` class for ramp settings.

    Parameters
    ----------------
    wind_mode : str
        Window mode type 'FULL', 'STRIPE', 'WINDOW'.
    xpix : int
        Size of window in x-pixels for frame time calculation.
    ypix : int
        Size of window in y-pixels for frame time calculation.
    x0 : int
        Lower-left x-coord position of detector window.
    y0 : int
        Lower-left y-coord position of detector window.

    Keyword Args
    ------------
    read_mode : str
        NIRCam Ramp Readout mode such as 'RAPID', 'BRIGHT1', etc.
    nint : int
        Number of integrations (ramps).
    ngroup : int
        Number of groups in a integration.
    nf : int
        Number of frames per group.
    nd1 : int
        Number of drop frame after reset (before first group read). 
    nd2 : int
        Number of drop frames within a group (ie., groupgap). 
    nd3 : int
        Number of drop frames after final read frame in ramp. 


    Examples
    --------
    Use kwargs functionality to pass keywords to the multiaccum class.
    
    Send via a dictionary of keywords and values:
        >>> kwargs = {'read_mode':'RAPID', 'nint':5, 'ngroup':10}
        >>> d = det_timing(**kwargs)
    
    Set the keywords directly:   
        >>> d = det_timing(read_mode='RAPID', nint=5, ngroup=10)
    """

    def __init__(self, wind_mode='FULL', xpix=2048, ypix=2048, x0=0, y0=0, 
                 mode='SLOW', nff=None, **kwargs):

        self._opmode = mode.upper()
        
        if 'JWST' in self._opmode:
            pixrate = 1e5
            loh = 12
            nchans = 4
            reset_type = 'pixel'
        elif 'SLOW' in self._opmode:
            pixrate = 1e5
            loh = 8
            nchans = 32
            reset_type = 'pixel'
            if nff is None: nff = 0 
        elif 'MED' in self._opmode:
            pixrate = 3e5
            loh = 8
            nchans = 32
            reset_type = 'pixel'
            if nff is None: nff = 0 
        elif 'FAST' in self._opmode:
            pixrate = 5e6
            loh = 3
            nchans = 32
            reset_type = 'line'
            if nff is None: nff = 0 

        self._detector_pixels = 2048
        self._nchans = nchans
        self._nff = nff

        self.multiaccum = multiaccum(wind_mode=wind_mode, **kwargs)
        self.wind_mode = wind_mode.upper()
        self._xpix = xpix; self._x0 = x0
        self._ypix = ypix; self._y0 = y0

        self._validate_pixel_settings()

        # By default fast-scan readout direction is [-->,<--,-->,<--]
        # If same_scan_direction, then all --> 
        # If reverse_scan_direction, then [<--,-->,<--,-->] or all <--
        self.same_scan_direction = False
        self.reverse_scan_direction = False

        # Pixel Rate in Hz
        self._pixel_rate = pixrate
        # Number of extra clock ticks per line
        self._line_overhead = loh
        # Pixel or line resets
        self._reset_type = reset_type

    @property
    def wind_mode(self):
        """Window mode attribute"""
        return self._wind_mode
    @wind_mode.setter
    def wind_mode(self, value):
        """Set Window mode attribute"""
        self._wind_mode = value

    @property
    def y0(self):
        return int(self._y0)
    @y0.setter
    def y0(self, val):
        self._y0 = val
    @property
    def x0(self):
        return int(self._x0)
    @x0.setter
    def x0(self, val):
        self._x0 = val
    @property
    def ypix(self):
        return int(self._ypix)
    @ypix.setter
    def ypix(self, val):
        self._ypix = val
    @property
    def xpix(self):
        return int(self._xpix)
    @xpix.setter
    def xpix(self, val):
        self._xpix = val

    @property
    def nout(self):
        """Number of simultaneous detector output channels stripes"""
        return 1 if self.wind_mode == 'WINDOW' else self._nchans

    @property
    def chsize(self):
        """Size of Amplifier Channel"""
        return int(self.xpix // self.nout)

    @property
    def ref_info(self):
        """Array of reference pixel borders [lower, upper, left, right]."""
        det_size = self._detector_pixels
        x1 = self.x0; x2 = x1 + self.xpix
        y1 = self.y0; y2 = y1 + self.ypix

        w = 4 # Width of ref pixel border
        lower = int(w-y1)
        upper = int(w-(det_size-y2))
        left  = int(w-x1)
        right = int(w-(det_size-x2))
        ref_all = np.array([lower,upper,left,right], dtype='int')
        ref_all[ref_all<0] = 0
        return ref_all

    @property
    def mask_act(self):
        """Active pixel mask for det coordinates"""
        # mask_act = np.zeros([self.ypix,self.xpix]).astype('bool')
        # rb, rt, rl, rr = self.ref_info
        # mask_act[rb:-rt,rl:-rr] = True  # This doesn't work if rr or rt are 0!!
        return ~self.mask_ref
    @property
    def mask_ref(self):
        """Reference pixel mask for det coordinates"""
        # [bottom, upper, left, right]
        rb, rt, rl, rr = self.ref_info
        ref_mask = np.zeros([self.ypix,self.xpix], dtype=bool)
        if rb>0: ref_mask[0:rb,:] = True
        if rt>0: ref_mask[-rt:,:] = True
        if rl>0: ref_mask[:,0:rl] = True
        if rr>0: ref_mask[:,-rr:] = True
        return ref_mask
    @property
    def mask_channels(self):
        """Channel masks for det coordinates"""
        ch_mask = np.zeros([self.ypix,self.xpix])
        for ch in np.arange(self.nout):
            ch_mask[:,ch*self.chsize:(ch+1)*self.chsize] = ch
        return ch_mask

    @property
    def nff(self):
        """Number of fast row resets that occur before Reset Frame"""
        if self._nff is None:
            if self.wind_mode=='WINDOW': 
                if   self.ypix>256: nff = 2048
                elif self.ypix>64:  nff = 512
                elif self.ypix>16:  nff = 256
                elif self.ypix>8:   nff = 64
                else:               nff = 16
            elif self.wind_mode=='STRIPE': 
                if   self.ypix==2048: nff = 0
                elif self.ypix>=256:  nff = 2048
                else:                 nff = 512
            elif self.wind_mode=='FULL':
                nff = 0
        else:
            nff = self._nff
        
        return nff
    @nff.setter
    def nff(self, val):
        self._nff = val

    def _validate_pixel_settings(self):
        """ 
        Validation to make sure the defined pixel sizes are consistent with
        detector readout mode (FULL, STRIPE, or WINDOW)
        """

        wind_mode = self.wind_mode

        modes = ['FULL', 'STRIPE', 'WINDOW']
        if wind_mode not in modes:
            wstr = str.join(', ', modes)
            raise ValueError("{} not a valid readout mode. Acceptable values: {}".format(wind_mode, wstr))

        detpix = self._detector_pixels
        xpix = self.xpix; x0 = self.x0
        ypix = self.ypix; y0 = self.y0

        # Check some consistencies with frame sizes
        if wind_mode == 'FULL':
            if ypix != detpix:
                _log.warning(f'In {wind_mode} mode, but ypix not {detpix}. Setting ypix={detpix}.')
                ypix = detpix
            if y0 != 0:
                _log.warning(f'In {wind_mode} mode, but x0 not 0. Setting y0=0.')
                y0 = 0

        if (wind_mode == 'STRIPE') or (wind_mode == 'FULL'):
            if xpix != detpix:
                _log.warning(f'In {wind_mode} mode, but xpix not {detpix}. Setting xpix={detpix}.')
                xpix = detpix
            if x0 != 0:
                _log.warning(f'In {wind_mode} mode, but x0 not 0. Setting x0=0.')
                x0 = 0
    
        if (x0+xpix) > detpix:
            raise ValueError("x0+xpix ({}+{}) is larger than detector size ({})!"\
                             .format(x0,xpix,detpix))
        if (y0+ypix) > detpix:
            raise ValueError("y0+ypix ({}+{}) is larger than detector size ({})!"\
                             .format(y0,ypix,detpix))

        # Update values if no errors were thrown
        self._xpix = xpix; self._x0 = x0
        self._ypix = ypix; self._y0 = y0

    def _fix_precision(self, input):
        """
        Many timing calculations result from minor precision issues with very
        small numbers (1e-16) added to the real result. This function attempts
        to truncate these small innaccuracies by dividing by the clock sample
        time to get the total integer number of clock cycles.
        """
        return int(input * self._pixel_rate + 0.5) / self._pixel_rate

    @property
    def _extra_lines(self):
        """Number of extra lines per frame.
        
        Determine how many extra lines/rows are added to a to a given frame.
        Based on ASIC microcode 10 and NIRCam arrays. Other instruments might
        operate differently.
        """
        if self.nout == 1:
            xtra_lines = 2 if self.xpix>10 else 3
        else:
            xtra_lines = 1
            
        return xtra_lines
        
    @property
    def _exp_delay(self):
        """Transition to idle delay (sec)
        
        Additional overhead time at the end of an exposure.
        This does not add any more photon flux to a pixel.
        Due to transition to idle.
        """

        # Window Mode        
        if self.nout == 1:
            if   self.xpix>150: xtra_lines = 0
            elif self.xpix>64:  xtra_lines = 1
            elif self.xpix>16:  xtra_lines = 2
            elif self.xpix>8:   xtra_lines = 4
            else:               xtra_lines = 5
        # Full and Stripe
        else: 
            xtra_lines = 1
        
        # Clock ticks per line
        xticks = self.chsize + self._line_overhead
        
        # Return number of seconds
        return xticks * xtra_lines / self._pixel_rate


    @property
    def _frame_overhead_pix(self):
        """
        Full and Stripe mode frames have an additional pixel at the end.
        """
        
        pix_offset = 0 if self.nout==1 else 1
        return pix_offset
        
    @property
    def time_row_reset(self):
        """NFF Row Resets time per integration"""
        
        nff = self.nff
            
        xtra_lines = int(nff / (self.chsize))
        
        # Clock ticks per line
        xticks = self.chsize + self._line_overhead  
        return xticks * xtra_lines / self._pixel_rate
        
    @property
    def time_frame(self):
        """Determine frame time (sec) based on xpix, ypix, and wind_mode."""

        chsize = self.chsize                   # Number of x-pixels within a channel
        xticks = chsize + self._line_overhead  # Clock ticks per line
        flines = self.ypix + self._extra_lines # Lines per frame

        # Add a single pix offset for full frame and stripe.
        pix_offset = self._frame_overhead_pix
        #end_delay = 0 # Used for syncing each frame w/ FPE bg activity. Not currently used.

        # Total number of clock ticks per frame (reset, read, and drops)
        fticks = xticks*flines + pix_offset

        # Return frame time
        return fticks / self._pixel_rate

    @property
    def time_group(self):
        """Time per group based on time_frame, nf, and nd2."""
        return self.time_frame * (self.multiaccum.nf + self.multiaccum.nd2)

    @property
    def time_ramp(self):
        """Photon collection time for a single ramp."""

        # How many total frames (incl. dropped and all) per ramp?
        # Exclude nd3 (drops that add nothing)
        ma = self.multiaccum
        nf = ma.nf; nd1 = ma.nd1; nd2 = ma.nd2
        ngroup = ma.ngroup

        tint = (nd1 + ngroup*nf + (ngroup-1)*nd2) * self.time_frame
        return self._fix_precision(tint)

    @property
    def time_int(self):
        """Same as time_ramp, except that 'int' follows the JWST nomenclature"""
        return self.time_ramp

    @property
    def time_ramp_eff(self):
        """Effective ramp time for slope fit tf*(ng-1)"""
        
        ma = self.multiaccum
        if ma.ngroup<=1:
            res = self.time_frame * (ma.nd1 + (ma.nf + 1) / 2)
        else:
            res = self.time_group * (ma.ngroup - 1)

        return self._fix_precision(res)

    @property
    def time_int_eff(self):
        """Same as time_ramp_eff, except that 'int' follows the JWST nomenclature"""
        return self.time_ramp_eff

    @property
    def time_exp(self):
        """Total photon collection time for all ramps."""
        res = self.multiaccum.nint * self.time_ramp
        return self._fix_precision(res)

    # @property
    # def time_total_int(self):
    #     """Total time for all frames in a ramp.
        
    #     Includes resets and excess drops, as well as NFF Rows Reset.
    #     """

    #     ma = self.multiaccum
    #     nf = ma.nf; nd1 = ma.nd1; nd2 = ma.nd2; nd3 = ma.nd3
    #     ngroup = ma.ngroup
    #     nr = 1

    #     nframes = nr + nd1 + ngroup*nf + (ngroup-1)*nd2 + nd3        
    #     return nframes * self.time_frame + self.time_row_reset

    @property
    def time_total_int1(self):
        """Total time for all frames in first ramp of exposure.
        
        Includes resets and excess drops, as well as NFF Rows Reset.
        """

        ma = self.multiaccum
        nf = ma.nf; nd1 = ma.nd1; nd2 = ma.nd2; nd3 = ma.nd3
        ngroup = ma.ngroup
        nr = ma.nr1

        nframes = nr + nd1 + ngroup*nf + (ngroup-1)*nd2 + nd3        
        res = nframes * self.time_frame + self.time_row_reset
        return self._fix_precision(res)

    @property
    def time_total_int2(self):
        """Total time for all frames in a subsequent ramp.
        
        Includes resets and excess drops, as well as NFF Rows Reset.
        Only differs from time_total_int1 in case nr1 != nr2
        """

        ma = self.multiaccum
        nf = ma.nf; nd1 = ma.nd1; nd2 = ma.nd2; nd3 = ma.nd3
        ngroup = ma.ngroup
        nr = ma.nr2
        
        # Set to 0 if only a single integration
        if ma.nint <= 1:
            return 0.
        else:
            nframes = nr + nd1 + ngroup*nf + (ngroup-1)*nd2 + nd3        
            res = nframes * self.time_frame + self.time_row_reset
            return self._fix_precision(res)

    @property
    def time_total(self):
        """Total exposure acquisition time"""
        # exp1 = 0 if self.multiaccum.nint == 0 else self.time_total_int1
        # exp2 = 0 if self.multiaccum.nint <= 1 else self.time_total_int2 * (self.multiaccum.nint-1)
        exp1 = self.time_total_int1
        exp2 = self.time_total_int2 * (self.multiaccum.nint-1)
        res = exp1 + exp2 + self._exp_delay
        return self._fix_precision(res)

    @property
    def times_group_avg(self):
        """Times at each averaged group since reset"""
        ma = self.multiaccum
        nf_avg = np.arange(ma.nf+1).sum() / ma.nf
        return np.arange(ma.ngroup) * self.time_group + (ma.nd1 + nf_avg) * self.time_frame

    def to_dict(self, verbose=False):
        """Export detector settings to a dictionary."""

        p = [('wind_mode',self.wind_mode), ('nout',self.nout), 
             ('xpix',self.xpix), ('ypix',self.ypix), ('x0',self.x0), ('y0',self.y0)]
        return tuples_to_dict(p, verbose)

    def times_to_dict(self, verbose=False):
        """Export ramp times as dictionary with option to print output to terminal."""

        times = [('t_frame', self.time_frame), ('t_group', self.time_group), 
                 ('t_int',   self.time_int),   ('t_exp',   self.time_exp), 
                 ('t_acq',   self.time_total), 
                 ('t_int_tot1', self.time_total_int1), 
                 ('t_int_tot2', self.time_total_int2)]
        return tuples_to_dict(times, verbose)

    def int_times_table(self, date_start, time_start, offset_seconds=None):
        """Create and populate the INT_TIMES table, which is saved as a
        separate extension in the output data file.

        Parameters
        ----------
        date_start : str
            Date string of observation ('2020-02-28')
        time_start : str
            Time string of observation ('12:24:56')
        offset_seconds : None or float
            Time from beginning of observation until start of integration.

        Returns
        -------
        int_times_tab : astropy.table.Table
            Table of starting, mid, and end times for each integration
        """
        
        from astropy.table import Table
        from astropy.time import Time, TimeDelta
        from astropy import units as u

        if offset_seconds is None:
            offset_seconds = 0

        integration_numbers = np.arange(self.multiaccum.nint)

        start_time_string = date_start + 'T' + time_start
        start_time = Time(start_time_string) + offset_seconds * u.second

        integration_time = self.time_total_int2
        integ_time_delta = TimeDelta(integration_time * u.second)
        start_times = start_time + (integ_time_delta * integration_numbers)

        reset_time = self.multiaccum.nr2 * self.time_frame
        integration_time_exclude_reset = TimeDelta((integration_time - reset_time) * u.second)
        end_times = start_times + integration_time_exclude_reset

        mid_times = start_times + integration_time_exclude_reset / 2.

        # For now, let's keep the BJD (Barycentric?) times identical
        # to the MJD times.
        start_times_bjd = start_times
        mid_times_bjd = mid_times
        end_times_bjd = end_times

        # Create table
        nrows = len(integration_numbers)
        data_list = [(integration_numbers[i] + 1, 
                      start_times.mjd[i], mid_times.mjd[i], end_times.mjd[i],
                      start_times_bjd.mjd[i], mid_times_bjd.mjd[i], end_times_bjd.mjd[i]) 
                     for i in range(nrows)]

        int_times_tab = np.array(data_list,
                                 dtype=[('integration_number','<i2'),
                                        ('int_start_MJD_UTC','<f8'),
                                        ('int_mid_MJD_UTC', '<f8'),
                                        ('int_end_MJD_UTC','<f8'),
                                        ('int_start_BJD_TDB','<f8'),
                                        ('int_mid_BJD_TDB','<f8'),
                                        ('int_end_BJD_TDB','<f8')])

        return int_times_tab
        
    def pixel_noise(self, ng=None, nf=None, verbose=False, **kwargs):
        """Noise values per pixel.
        
        Return theoretical noise calculation for the specified MULTIACCUM exposure 
        in terms of e-/sec. This uses the pre-defined detector-specific noise 
        properties. Can specify flux of a source as well as background and 
        zodiacal light (in e-/sec/pix). After getting the noise per pixel per
        ramp (integration), value(s) are divided by the sqrt(NINT) to return
        the final noise

        Parameters
        ----------
        ng : None or int or image
            Option to explicitly state number of groups. This is specifically
            used to enable the ability of only calculating pixel noise for
            unsaturated groups for each pixel. If a numpy array, then it should
            be the same shape as `fsrc` image. By default will use `self.multiaccum.ngroup`.
        nf : int
            Option to explicitly states number of frames in each group.
            By default will use `self.multiaccum.nf`.
        verbose : bool
            Print out results at the end.

        Keyword Arguments
        -----------------
        rn : float
            Read Noise per pixel (e-).
        ktc : float
            kTC noise (in e-). Only valid for single frame (n=1)
        p_excess : array-like
            An array or list of two elements that holds the parameters
            describing the excess variance observed in effective noise plots.
            By default these are both 0. For NIRCam detectors, recommended
            values are [1.0,5.0] for SW and [1.5,10.0] for LW.
        idark : float
            Dark current in e-/sec/pix.
        fsrc : float
            Flux of source in e-/sec/pix.
        fzodi : float
            Zodiacal light emission in e-/sec/pix.
        fbg : float
            Any additional background (telescope emission or scattered light?)
        ideal_Poisson : bool
            If set to True, use total signal for noise estimate,
            otherwise MULTIACCUM equation is used.

        Notes
        -----
        fsrc, fzodi, and fbg are functionally the same as they are immediately summed.
        They can also be single values or multiple elements (list, array, tuple, etc.).
        If multiple inputs are arrays, make sure their array sizes match.
        
        """

        ma = self.multiaccum
        if ng is None:
            ng = ma.ngroup
        if nf is None:
            nf = ma.nf

        # Pixel noise per ramp (e-/sec/pix)
        pn = pix_noise(ngroup=ng, nf=nf, nd2=ma.nd2, tf=self.time_frame, **kwargs)
    
        # Divide by sqrt(Total Integrations)
        final = pn / np.sqrt(ma.nint)
        if verbose:
            print('Noise (e-/sec/pix): {}'.format(final))
            print('Total Noise (e-/pix): {}'.format(final*self.time_exp))

        return final

    def pix_timing_map(self, same_scan_direction=None, reverse_scan_direction=None,
                       avg_groups=False, reset_zero=False, return_flat=False):
        """Create array of pixel times for a single ramp. 
        
        Each pixel value corresponds to the precise time at which
        that pixel was read out during the ramp acquisition. The first
        pixel(s) have t=0.
        
        Parameters
        ----------
        same_scan_direction : bool
            Are all the output channels read in the same direction?
            By default fast-scan readout direction is ``[-->,<--,-->,<--]``
            If ``same_scan_direction``, then all ``-->``
        reverse_scan_direction : bool
            If ``reverse_scan_direction``, then ``[<--,-->,<--,-->]`` or all ``<--``
        avg_groups : bool
            For groups where nf>1, the telescope data gets averaged via a 
            bit-shifter. Setting ``avg_groups=True`` also averages the
            pixel times in a similar manner. Default is False.
        return_flat : bool
            Return a single array rather than image array.
        
        Keyword Args
        ------------
        reset_zero : bool
            Return timing relative to when reset to get photon-collection time of each pixel.
            Otherwise, t=0 corresponds to very first pixel(s) read in the ramp.
            
        Returns
        -------
        ndarray
            If ``return_flat=True`` then the data is a flattened array for a
            single channel output. Otherwise, the output is a data cube of the
            same size and shape of the raw data with these detector settings.

        Example
        -------
        Assume you have a cube of raw full frame data (RAPID, ngroup=5).
        Create a det_timing instance and get channel:
        
        >>> d = det_timing(ngroup=5)
        >>> tarr = d.pixel_timing_map(return_flat=True, avg_groups=True)
        
        >>> nx, ny = (d.xpix, d.ypix)
        >>> nout   = d.nout      # Number of amplifier output channels
        >>> chsize = d.chsize    # Channel size (x-direction)
        >>> # Reshape into (nz, ny, nout, chsize)
        >>> data = data.reshape([-1,ny,nout,chsize])
        >>> # Reverse odd channels in x-direction to match even chans
        >>> for ch in range(nout):
        >>>     if np.mod(ch,2)==1:
        >>>         data[:,:,ch,:] = data[:,:,ch,::-1]
        >>> # Final data reshaped into 4 flattened output channels
        >>> data = data.transpose([0,1,3,2]).reshape([-1,nout])
        >>> # Can plot this like plt.plot(tarr, data) to make nout line plots
        
        """
        
        xpix = self.xpix
        ypix = self.ypix
        nout = self.nout

        chsize = self.chsize                   # Number of x-pixels within a channel
        xticks = chsize + self._line_overhead  # Clock ticks per line
        flines = ypix + self._extra_lines      # Lines per frame

        if same_scan_direction is None:
            same_scan_direction = self.same_scan_direction
        if reverse_scan_direction is None:
            reverse_scan_direction = self.reverse_scan_direction
        
        # Pixel-by-pixel or line-by-line reset?
        line_reset = True if 'line' in self._reset_type.lower() else False

        # Add a single pix offset for full frame and stripe.
        pix_offset = self._frame_overhead_pix

        # Total number of clock ticks per frame (reset, read, and drops)
        fticks = xticks*flines + pix_offset

        # Total ticks in a ramp (exclude nd3 drop frames)
        ma = self.multiaccum
        nf = ma.nf; nd1 = ma.nd1; nd2 = ma.nd2
        ngroup = ma.ngroup

        nframes = nd1 + ngroup*nf + (ngroup-1)*nd2
        nticks = fticks * nframes
        
        # Make large array of pixel timing
        arr = np.arange(nticks, dtype=float).reshape([nframes,-1])
        arr_reset = np.arange(fticks, dtype=float)
        if reset_zero:
            arr += arr_reset.max() + 1
        
        # Chop off single pix overhead
        if pix_offset>0:
            arr = arr[:,:-pix_offset]
            arr_reset = arr_reset[:-pix_offset]
            
        # Reshape to 3D array
        arr = arr.reshape([nframes, flines, xticks])
        arr_reset = arr_reset.reshape([flines, xticks])

        # If performing line-by-line reset, all pixels in a line 
        # were reset at the same time.
        if line_reset:
            for i in range(flines):
                arr_reset[i,:] = np.max(arr_reset[i,:])
        
        # Chop off x pixel & y line overheads
        arr = arr[:,:ypix,:chsize]
        arr_reset = arr_reset[:ypix,:chsize]
        
        # By default fast-scan readout direction is [-->,<--,-->,<--]
        # If same_scan_direction, then all --> 
        # If reverse_scan_direction, then [<--,-->,<--,-->] or all <--
        if reverse_scan_direction:
            arr = arr[::-1]
            arr_reset = arr_reset[:,::-1]

        arr_list = []
        arr2_list = []
        if nout>1:
            # Consecutive outputs reversed?
            for ch in range(nout):
                if (np.mod(ch,2) == 0) or (same_scan_direction == True): 
                    arr_list.append(arr)
                    arr2_list.append(arr_reset)
                else: 
                    arr_list.append(arr[:,:,::-1])
                    arr2_list.append(arr_reset[:,::-1])
            data = np.concatenate(arr_list, axis=2)
            data_reset = np.concatenate(arr2_list, axis=1)
        else:
            data = arr
            data_reset = arr_reset
            
        del arr, arr_list, arr2_list, arr_reset   

        # Timing for averaged (bit-shifted) frames
        # Remove drops and average grouped data
        if (avg_groups and nf>1) or (nd2>0):
            # Trailing drop frames already excluded
            # so need to pull off last group of frames
            # in order to properly reshape things.
            if avg_groups and (nf>1):
                data_end = data[-nf:,:,:].mean(axis=0)
            else:
                data_end = data[-nf:,:,:]
            data_end = data_end.reshape([-1,ypix,xpix])

            # Only care about first (n-1) groups
            # Last group is handled separately
            # Cut off last set of nf frames
            data = data[:-nf,:,:]

            # Reshape for easy group manipulation
            data = data.reshape([-1,nf+nd2,ypix,xpix])

            # Trim off the dropped frames (nd2)
            if nd2>0: data = data[:,:nf,:,:]

            # Average the frames within groups
            # In reality, the 16-bit data is bit-shifted
            if avg_groups and (nf>1):
                data = data.mean(axis=1)
            else:
                data = data.reshape([-1,ypix,xpix])

            # Add back the last group (already averaged)
            data = np.append(data,data_end,axis=0)

        if reset_zero:
            for im in data:
                im -= data_reset
            
        # Put into time
        # print(data.dtype)
        data /= self._pixel_rate

        # Return timing info
        if return_flat: # Flatten array
            return data.ravel()
        else: # Get rid of dimensions of length 1
            return data.squeeze()


def tuples_to_dict(pairs, verbose=False):
    """
    Take a list of paired tuples and convert to a dictionary
    where the first element of each tuple is the key and the 
    second element is the value.
    """
    d={}
    for (k, v) in pairs:
        d[k] = v
        if verbose:
            if isinstance(v,float): print("{:<10} {:>10.4f}".format(k, v))
            else: print("{:<10} {:>10}".format(k, v))
    return d


def nrc_header(det_class, filter=None, pupil=None, obs_time=None, header=None,
               DMS=True, targ_name=None):
    """Simulated header

    Create a generic NIRCam FITS header from a detector_ops class.

    Parameters
    ----------
    filter : str
        Name of filter element.
    pupil : str
        Name of pupil element.
    DMS : bool
        Make the header in a format used by Data Management Systems.
    obs_time : datetime
        Specifies when the observation was considered to be executed.
        If not specified, then it will choose the current time.
        This must be a datetime object:

            >>> datetime.datetime(2016, 5, 9, 11, 57, 5, 796686)

    header : obj
        Can pass an existing header that will be updated.
        This has not been fully tested.
    targ_name : str
        Standard astronomical catalog name for a target.
        Otherwise, it will be UNKNOWN.
    """

    from .version import __version__

    filter = 'UNKNOWN' if filter is None else filter
    pupil  = 'UNKNOWN' if pupil  is None else pupil
    targ_name = 'UNKNOWN' if targ_name is None else targ_name

    d = det_class
    # MULTIACCUM ramp information
    ma = d.multiaccum

    # How many axes?
    naxis = 2 if ma.ngroup == 1 else 3
    if naxis == 3:
        naxis3 = ma.ngroup
    naxis1 = d.xpix
    naxis2 = d.ypix

    # Select Detector ID based on SCA ID
    detector = d.detname

    # Are we in subarray?
    sub_bool = True if d.wind_mode != 'FULL' else False
    # Horizontal window mode?
    hwinmode = 'ENABLE' if d.wind_mode=='WINDOW' else 'DISABLE'

    # Window indices (0-indexed)
    x1 = d.x0; x2 = x1 + d.xpix
    y1 = d.y0; y2 = y1 + d.ypix

    # Ref pixel info
    ref_all = d.ref_info

    # Dates and times
    obs_time = datetime.datetime.utcnow() if obs_time is None else obs_time
    # Total time to complete obs including all overheads
    tdel = d.time_total 
    dtstart = obs_time.isoformat()
    aTstart = Time(dtstart)
    dtend = (obs_time + datetime.timedelta(seconds=tdel)).isoformat()
    aTend = Time(dtend)
    dstart = dtstart[:10]; dend = dtend[:10]
    tstart = dtstart[11:-3]; tend = dtend[11:-3]
    tsample = 1e6/d._pixel_rate

    ################################################################
    # Create blank header
    hdr_update = False  if header is None else True
    hdr = fits.Header() if header is None else header

    # Add in basic header info
    hdr['SIMPLE']  = (True,   'conforms to FITS standard')
    hdr['BITPIX']  = (16,     'array data type')
    if DMS == True:
        hdr['SUBSTRT1'] = (x1+1, 'Starting pixel in axis 1 direction')
        hdr['SUBSTRT2'] = (y1+1, 'Starting pixel in axis 2 direction')
        hdr['SUBSIZE1'] = naxis1
        hdr['SUBSIZE2'] = naxis2
        hdr['NAXIS'] = (naxis,  'number of array dimensions')
    else:
        hdr['NAXIS']   = (naxis,  'number of array dimensions')
        hdr['NAXIS1']  = naxis1
        hdr['NAXIS2']  = naxis2

        if hdr_update: hdr.pop('NAXIS3', None)
        if naxis == 3: hdr['NAXIS3']  = (naxis3, 'length of third data axis')
    hdr['EXTEND']  = True

    hdr['DATE']    = ('',   'date file created (yyyy-mm-ddThh:mm:ss,UTC)')
    hdr['BSCALE']  = (1,     'scale factor for array value to physical value')
    hdr['BZERO']   = (32768, 'physical value for an array value of zero')
    hdr['UNITS']   = ('',  'Units for the data type (ADU, e-, etc.)')
    hdr['ORIGIN']  = ('UAz',  'institution responsible for creating FITS file')
    hdr['FILENAME']= ('',   'name of file')
    hdr['FILETYPE']= ('raw', 'type of data found in data file')

    # Observation Description
    hdr['TELESCOP']= ('JWST',    'telescope used to acquire data')
    hdr['INSTRUME']= ('NIRCAM',  'instrument identifier used to acquire data')
    hdr['OBSERVER']= ('UNKNOWN', 'person responsible for acquiring data')
    hdr['DATE-OBS']= (dstart, 'UT date of observation (yyyy-mm-dd)')
    hdr['TIME-OBS']= (tstart, 'Approximate UT time of start of observation (hh:mm:ss.sss)')
    if DMS == True:
        if 'GRISM' in pupil:
            exp_type = 'NRC_GRISM'
        elif pupil == None:
            exp_type = 'UNKNOWN'
        else:
            exp_type = 'NRC_IMAGE'
        hdr['EXP_TYPE'] = (exp_type,'Type of data in the exposure')
    hdr['DATE-END']= (dend,   'UT date of end of observation(yyyy-mm-dd)')
    hdr['TIME-END']= (tend,   'UT time of end of observation (hh:mm:ss.sss)')
    hdr['SCA_ID']  = (d.scaid,   'Unique SCA identification in ISIM')
    hdr['DETECTOR']= (d.detname, 'ASCII Mnemonic corresponding to the SCA_ID')
    hdr['PIXELSCL']= (d.pixelscale, 'Detector Pixel Scale (arcsec/pixel)')

    nx_noref = naxis1 - ref_all[2] - ref_all[3]
    ny_noref = naxis2 - ref_all[0] - ref_all[1]
    fovx = nx_noref * d.pixelscale
    fovy = ny_noref * d.pixelscale
    hdr['FOV']     = ('{:.2f}x{:.2f}'.format(fovx,fovy), 'Field of view in arcsec')

    if DMS == True:
        hdr['TARG_RA']=  (0.0, 'Target RA at mid time of exposure') #arbitrary position
        hdr['TARG_DEC']= (0.0, 'Target Dec at mid time of exposure') #arbitrary position

        hdr['PROGRAM'] = ('12345', 'Program number')
        hdr['OBSERVTN']= ('001',   'Observation number')
        hdr['VISIT']   = ('001',   'Visit Number')
        hdr['VISITGRP']= ('01',  'Visit Group Identifier')

        hdr['SEQ_ID']  = ('1', 'Parallel sequence identifier')
        hdr['ACT_ID']  = ('1', 'Activity identifier')
        hdr['EXPOSURE']= ('1', 'Exposure request number')
        hdr['OBSLABEL']= ('Target 1 NIRCam Observation 1', 'Proposer label for the observation')
        hdr['EXPSTART']= (aTstart.mjd, 'UTC exposure start time')
        hdr['EXPEND']  = (aTend.mjd, 'UTC exposure end time')
        hdr['EFFEXPTM']= (d.time_int_eff*d.time_int, 'Effective exposure time (sec)')
        hdr['NUMDTHPT']= ('1','Total number of points in pattern')
        hdr['PATT_NUM']= (1,'Position number in primary pattern')

    hdr['TARGNAME'] = (targ_name, 'Standard astronomical catalog name for target')
    hdr['OBSMODE'] = ('UNKNOWN', 'Observation mode')

    if DMS == True:
        if d.channel == 'LW':
            headerChannel = 'LONG'
        elif d.channel == 'SW':
            headerChannel = 'SHORT'
        else:
            headerChannel = 'UNKNOWN'
        hdr['CHANNEL'] = headerChannel

        hdr['GRATING'] = ('N/A - NIRCam', 'Name of the grating element used')
        hdr['BAND']    = ('N/A - NIRCam', 'MRS wavelength band')
        hdr['LAMP']    = ('N/A - NIRCam', 'Internal lamp state')
        hdr['GWA_XTIL']= ('N/A - NIRCam', 'Grating X tilt angle relative to mirror')
        hdr['GWA_YTIL']= ('N/A - NIRCam', 'Grating Y tilt angle relative to mirror')
        hdr['GWA_TILT']= ('N/A - NIRCam', 'GWA TILT (avg/calib) temperature (K)')
        hdr['MSAMETFL']= ('N/A - NIRCam', 'MSA metadata file name')
        hdr['MSAMETID']= ('N/A - NIRCam', 'MSA metadata ID')

    # Positions of optical elements
    hdr['FILTER']  = (filter, 'Module ' + d.module + ' ' + d.channel + ' FW element')
    hdr['PUPIL']   = (pupil, 'Module ' + d.module + ' ' + d.channel + ' PW element')
    hdr['PILSTATE']= ('RETRACTED', 'Module ' + d.module + ' PIL deploy state')

    # Readout Mode
    hdr['NSAMPLE'] = (1,            'A/D samples per read of a pixel')
    if DMS == True:
        frmName = 'NFRAMES'
        grpName = 'NGROUPS'
        intName = 'NINTS'
    else:
        frmName = 'NFRAME'
        grpName = 'NGROUP'
        intName = 'NINT'
    hdr[frmName]   = (ma.nf,         'Number of frames in group')
    hdr[grpName]   = (ma.ngroup,     'Number groups in an integration')
    hdr[intName]   = (ma.nint,     'Number of integrations in an exposure')

    # Timing information
    hdr['TSAMPLE'] = (tsample,        'Delta time between samples in microsec')
    hdr['TFRAME']  = (d.time_frame,   'Time in seconds between frames')
    hdr['TGROUP']  = (d.time_group,   'Delta time between groups')
    hdr['DRPFRMS1']= (ma.nd1, 'Number of frame skipped prior to first integration')
    hdr['GROUPGAP']= (ma.nd2, 'Number of frames skipped')
    hdr['DRPFRMS3']= (ma.nd3, 'Number of frames skipped between integrations')
    hdr['FRMDIVSR']= (ma.nf,  'Divisor applied to each group image')
    hdr['INTAVG']  = (1, 'Number of integrations averaged in one image')
    hdr['NRESETS1']= (ma.nr1, 'Number of reset frames prior to first integration')
    hdr['NRESETS2']= (ma.nr2, 'Number of reset frames between each integration')
    hdr['INTTIME'] = (d.time_int, 'Total integration time for one MULTIACCUM')
    hdr['EXPTIME'] = (d.time_exp, 'Exposure duration (seconds) calculated')
    hdr['FASTAXIS']= (d.fastaxis, 'Fast readout direction relative to image axes for Amp1')
    hdr['SLOWAXIS']= (d.slowaxis, 'Slow readout direction relative to image axes')
    
    # Subarray names
    if DMS == True:
        if (d.xpix == 2048) & (d.ypix == 2048):
            subName = 'FULL'
        elif (d.xpix == 640) & (d.ypix == 640):
            subName = 'SUB640'
        elif (d.xpix == 320) & (d.ypix == 320):
            subName = 'SUB320'
        elif (d.xpix == 400) & (d.ypix == 400):
            subName = 'SUB400P'
        elif (d.xpix == 64) & (d.ypix == 64):
            subName = 'SUB64P'
        elif (d.xpix == 2048) & (d.ypix == 256):
            subName = 'SUBGRISM256'
        elif (d.xpix == 2048) & (d.ypix == 128):
            subName = 'SUBGRISM128'
        elif (d.xpix == 2048) & (d.ypix == 64):
            subName = 'SUBGRISM64'
        else:
            subName = 'UNKNOWN'
        hdr['SUBARRAY']= (subName, 'Detector subarray string')
    else:
        hdr['SUBARRAY']= (sub_bool, 'T if subarray used, F if not')
        hdr['HWINMODE']= (hwinmode, 'If enabled, single output mode used')

    # Readout Patterns
    if DMS == True:
        hdr['READPATT']= (ma.read_mode, 'Readout pattern name')
        hdr['ZROFRAME']= (True,       'T if zeroth frame present, F if not')
    else:
        hdr['READOUT'] = (ma.read_mode, 'Readout pattern name')
        hdr['ZROFRAME']= (False,       'T if zeroth frame present, F if not')

    #Reference Data
    hdr['TREFROW'] = (ref_all[1], 'top reference pixel rows')
    hdr['BREFROW'] = (ref_all[0], 'bottom reference pixel rows')
    hdr['LREFCOL'] = (ref_all[2], 'left col reference pixels')
    hdr['RREFCOL'] = (ref_all[3], 'right col reference pixels')
    hdr['NREFIMG'] = (0, 'number of reference rows added to end')
    hdr['NXREFIMG']= (0, 'reference image columns')
    hdr['NYREFIMG']= (0, 'reference image rows')
    hdr['COLCORNR']= (x1+1, 'The Starting Column for ' + detector)
    hdr['ROWCORNR']= (y1+1, 'The Starting Row for ' + detector)

    hdr.insert('EXTEND', '', after=True)
    hdr.insert('EXTEND', '', after=True)
    hdr.insert('EXTEND', '', after=True)

    hdr.insert('FILETYPE', '', after=True)
    hdr.insert('FILETYPE', ('','Observation Description'), after=True)
    hdr.insert('FILETYPE', '', after=True)

    hdr.insert('OBSMODE', '', after=True)
    hdr.insert('OBSMODE', ('','Optical Mechanisms'), after=True)
    hdr.insert('OBSMODE', '', after=True)

    hdr.insert('PILSTATE', '', after=True)
    hdr.insert('PILSTATE', ('','Readout Mode'), after=True)
    hdr.insert('PILSTATE', '', after=True)

    hdr.insert('ZROFRAME', '', after=True)
    hdr.insert('ZROFRAME', ('','Reference Data'), after=True)
    hdr.insert('ZROFRAME', '', after=True)

    hdr.insert('ROWCORNR', '', after=True)
    hdr.insert('ROWCORNR', '', after=True)

    hdr['comment'] = 'Simulated data generated by {} v{}'\
                      .format(__package__,__version__)

    return hdr

def config2(input, intype='int'):
    """NIRCam CONFIG2 (0x4011) Register

    Return a dictionary of configuration parameters depending on the
    value of CONFIG2 register (4011).

    Parameters
    ----------
    input : int, str
        Value of CONFIG2, nominally as an int. Binary and Hex values
        can also be passed as strings.
    intype: str
        Input type (int, hex, or bin) for integer, hex, string,
        or binary string.

    """
    if 'hex' in intype:
        if '0x' in input:
            input = int(input, 0)
        else:
            input = int(input, 16)
    if 'bin' in intype:
        if '0b' in input:
            input = int(input, 0)
        else:
            input = int(input, 2)

    # Convert to 16-bit binary string
    input = "{0:016b}".format(input)

    # Config2 Bits (Right to Left)
    # ----------------------------
    # 0 : Vertical Enable
    # 1 : Horizontal Enable
    # 2 : Global reset per integration
    # 3 : Enable Fast row-by-row reset (only in window/stripe)
    # 6-4 : Number of fast row resets per int
    # 7 : Window mode in Idle when window enabled?
    # 8 : 0 = Preamp reset per frame; 1 = reset per row
    # 9 : Permanent Reset
    # 10 : Single step mode
    # 11 : Test pattern
    # 12 : FGS window mode
    # 13 : Power down preamp, adc, and ap during Idle
    # 14 : Power down preamp, adc, and ap during Drop
    # 15 : 0 = Preamp reset per frame; 1 = reset per integration

    # NFF Rows Reset
    # --------------
    # 000 = 1
    # 001 = 4
    # 010 = 16
    # 011 = 64
    # 100 = 256
    # 101 = 512
    # 110 = 1024
    # 111 = 2048

    nff_dict = {'000':   1, '001':   4, '010':  16, '011':  64,
                '100': 256, '101': 512, '110':1024, '111':2048}

    # Reverse for easier indexing of single values
    input2 = input[::-1]

    d = {}
    d['00_window_vert']  = True if bool(int(input2[0])) else False
    d['01_window_horz']  = True if bool(int(input2[1])) else False
    d['02_global_reset'] = True if bool(int(input2[2])) else False
    d['03_rows_reset']   = True if bool(int(input2[3])) else False
    d['04_rows_nff']     = nff_dict.get(input2[4:7][::-1])
    d['07_idle_window']  = True if bool(int(input2[7])) else False
    d['08_pa_reset']     = 'row' if bool(int(input2[8])) else 'frame'
    d['09_perm_reset']   = True if bool(int(input2[9])) else False
    d['10_single_step']  = True if bool(int(input2[10])) else False
    d['11_test_patt']    = True if bool(int(input2[11])) else False
    d['12_fgs_wind']     = True if bool(int(input2[12])) else False
    d['13_power_idl']    = True if bool(int(input2[13])) else False
    d['14_power_drop']   = True if bool(int(input2[14])) else False
    d['15_pa_reset']     = 'int' if bool(int(input2[15])) else 'frame'

    return d

def create_detops(header, DMS=False, read_mode=None, nint=None, ngroup=None,
    detector=None, wind_mode=None, xpix=None, ypix=None, x0=None, y0=None,
    nff=None):
    """NIRCam Detector class from header

    Create a NIRCam detector class based on header settings.
    Can override settings with a variety of keyword arguments.

    Parameters
    ----------
    header : obj
        Header from NIRCam FITS file
    DMS : bool
        Is header format from Data Management Systems? Otherwises, ISIM-like.

    Keyword Args
    ------------
    read_mode : str
        NIRCam Ramp Readout mode such as 'RAPID', 'BRIGHT1', etc.
    nint : int
        Number of integrations (ramps).
    ngroup : int
        Number of groups in a integration.
    detector : int, str
        NIRCam detector ID (481-490) or SCA ID (A1-B5).
    wind_mode : str
        Window mode type 'FULL', 'STRIPE', 'WINDOW'.
    xpix : int
        Size of window in x-pixels for frame time calculation.
    ypix : int
        Size of window in y-pixels for frame time calculation.
    x0 : int
        Lower-left x-coord position of detector window.
    y0 : int
        Lower-left y-coord position of detector window.
    nff : int
        Number of fast row resets.

    """
    # Create detector class
    from pynrc.pynrc_core import DetectorOps

    # Detector ID
    if detector is None:
        detector = header.get('SCA_ID')
        if detector is None:
            detector = header.get('DETECTOR')         

    # Detector size
    xpix = header['SUBSIZE1'] if DMS else header['NAXIS1'] if xpix is None else xpix
    ypix = header['SUBSIZE2'] if DMS else header['NAXIS2'] if ypix is None else ypix

    # Subarray position
    # Headers are 1-indexed, while detector class is 0-indexed
    if x0 is None:
        x1 = header['SUBSTRT1'] if DMS else header['COLCORNR']
        x0 = x1 - 1
    if y0 is None:
        y1 = header['SUBSTRT2'] if DMS else header['ROWCORNR']
        y0 = y1 - 1

    # Subarray setting: Full, Stripe, or Window
    if wind_mode is None:
        if xpix==ypix==2048:
            wind_mode = 'FULL'
        else:
            log_prev = conf.logging_level
            setup_logging('ERROR', verbose=False)

            # Test if STRIPE or WINDOW
            det_stripe = DetectorOps(detector, 'STRIPE', xpix, ypix, x0, y0)
            det_window = DetectorOps(detector, 'WINDOW', xpix, ypix, x0, y0)
            dt_stripe = np.abs(header['TFRAME'] - det_stripe.time_frame)
            dt_window = np.abs(header['TFRAME'] - det_window.time_frame)
            wind_mode = 'STRIPE' if dt_stripe<dt_window else 'WINDOW'

            setup_logging(log_prev, verbose=False)


    # Add MultiAccum info
    hnames = ['READPATT', 'NINTS', 'NGROUPS'] if DMS else ['READOUT',  'NINT',  'NGROUP']

    read_mode = header[hnames[0]] if read_mode is None else read_mode
    nint      = header[hnames[1]] if nint      is None else nint
    ngroup    = header[hnames[2]] if ngroup    is None else ngroup

    ma_args = {'read_mode':read_mode, 'nint':nint, 'ngroup':ngroup}

    return DetectorOps(detector, wind_mode, xpix, ypix, x0, y0, nff, **ma_args)




#!/usr/bin/env python

# Import the usual libraries
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

import os, pickle
from astropy.io import fits

from webbpsf_ext.image_manip import frebin, pad_or_cut_to_size, rotate_offset
from webbpsf_ext.image_manip import fshift, fourier_imshift
from webbpsf_ext.maths import round_int, hist_indices, binned_statistic
from webbpsf_ext.coords import dist_image
from webbpsf_ext import robust

from tqdm import trange, tqdm

import logging
# Define logging
_log = logging.getLogger(__name__)
_log.setLevel(logging.INFO)


#################################################################
# KLIP and LOCI PSF Subtraction functions
#################################################################

def mkpsf_klip(bg_opt, sci_opt, nsvd=None, svd_frac=0.9):
    """ Create a PSF using KLIP PSF subtraction

    Parameters
    ----------
    bg_opt : ndarray
        Optimization region of reference observations ("background").
    sci_opt : ndarray
        Optimization region of science observations.
    nsvd : int, list, or ndarray
        Number of SVD modes to keep. If None, then the number of modes
        is determined by the fraction of cumulative eigenvalues.
    svd_frac : float
        Fraction of cumulative eigenvalues to keep. Default is 0.9.
    """
    
    try:
        _, s, VT = np.linalg.svd(bg_opt, full_matrices=False)
    except Exception as e:
        _log.info(f"Unexpected {e=}, {type(e)=}")
        _log.info(bg_opt)
        raise e
    
    # Eigenvalues
    if nsvd is None:
        # Eigenvalues
        evals = s**2
        evals /= evals.sum()
        evals_csum = evals.cumsum()

        # try:
        emin = evals_csum.min()
        # except ValueError:
        #     print(bg_opt.shape, evals)
        #     raise ValueError

        # Cut off at 90% cumulative power
        eval_cut = emin + svd_frac * (1.0 - emin) 
        ind = np.where(evals_csum < eval_cut)
        Z = VT[ind[0],:]
        
        res = np.dot(sci_opt, np.dot(Z.T,Z)) 
    else:
        nsvd = np.array([nsvd]).flatten()
        nres = len(nsvd)
        res = []
        for i in range(nres):
            Z = VT[0:nsvd[i],:]
            res.append(np.dot(sci_opt, np.dot(Z.T,Z)) )
        res = np.array(res)
    
    return res

def fit_loci_coeff(bg_opt, sci_opt, use_QR=True, use_old=False, 
                   nsvd=None, svd_frac=0.9, **kwargs):
    """Fit coefficients in optimization region for LOCI PSF subtraction
    
    Parameters
    ----------
    bg_opt : array_like
        Background optimization region.
    sci_opt : array_like
        Science optimization region (single vector size of bg_opt.shape[1]).
    use_QR : bool, optional
        Use QR decomposition to solve for coefficients. Only works for
        nsvd=None.
    use_old : bool, optional
        Use old method of solving for coefficients (not recommended).
        Only applicable for nsvd=None.
    nsvd : None, int, or array-like, optional
        Number of singular values to use in SVD. If None, then the number
        of singular values is chosen to be the minimum number that account 
        for svd_frac of the variance in the data.
    svd_frac : float, optional
        Fraction of variance to account for in SVD. Ignored if nsvd is not
        None.
    """
    
    import numpy.ma as ma

    # Deal with NaNs in bg array?
    inan_bg = np.isnan(bg_opt)
    if np.any(inan_bg):
        use_masked_bg = True
        bg_opt = ma.masked_invalid(bg_opt)
    else:
        use_masked_bg = False

    # Deal with NaNs in sci array?
    inan_sci = np.isnan(sci_opt)
    if np.any(inan_sci):
        use_masked_sci = True
        sci_opt = ma.masked_invalid(sci_opt)
    else:
        use_masked_sci = False

    nopt = bg_opt.shape[1]
    # Covariance Matrix
    try:
        A = ma.dot(bg_opt, bg_opt.T).data if use_masked_bg else np.dot(bg_opt, bg_opt.T)
        A /= (nopt-1.0)
    except:
        A = ma.cov(bg_opt).data if use_masked_bg else np.cov(bg_opt)

    # Data vector
    b_use_masked = use_masked_bg or use_masked_sci
    b = ma.dot(bg_opt, sci_opt.T).data if b_use_masked else np.dot(bg_opt, sci_opt.T)

    # Original SVD cut-off version
    if nsvd is not None:
        try:
            U,s,VT = np.linalg.svd(A, full_matrices=True)
        except Exception as e:
            print(f"Unexpected {e=}, {type(e)=}")
            print(A)
            raise e

        nsvd = np.array([nsvd]).flatten()
        nres = len(nsvd)
        coeff_all = []
        for ns in nsvd:
            # Pseudo-invert
            sinv = np.zeros_like(s)
            sinv[0:ns] = 1 / s[0:ns]
            Sinv = np.diag(sinv)
            Ainv = np.dot(VT.T, np.dot(Sinv, U.T)) / (nopt - 1.0)

            coeff = np.dot(Ainv, b)
            coeff_all.append(coeff)
        
        if nres == 1:
            coeff = coeff_all[0]
        else:
            coeff = np.array(coeff_all)

    elif use_old:
        try:
            U,s,VT = np.linalg.svd(A, full_matrices=True)
        except Exception as e:
            print(f"Unexpected {e=}, {type(e)=}")
            print(A)
            raise e
        
        # Eigenvalues
        evals = s**2
        evals /= evals.sum()
        evals_csum = evals.cumsum()
        emin = evals_csum.min()

        # Cut off at 90% cumulative power
        eval_cut = emin + svd_frac * (1.0 - emin) 
        ind_s = np.where(evals_csum < eval_cut)[0]

        # Pseudo-inversion of A matrix
        # with SVD cut-off
        sinv = np.zeros_like(s)
        sinv[ind_s] = 1 / s[ind_s]
        Sinv = np.diag(sinv)
        Ainv = np.dot(VT.T, np.dot(Sinv, U.T)) / (nopt - 1.0)

        # Coefficients fit to psf_lib that best fits science data
        # witin the optimization region
        coeff = np.dot(Ainv, b)

    elif use_QR:
        # QR decomposition of A matrix
        q, r = np.linalg.qr(A, 'reduced')
        # computing Q^T*b (project b onto the range of A)
        qTb = np.matmul(q.T, b)
        # solving R*x = Q^T*b
        res = np.linalg.lstsq(r, qTb, rcond=1-svd_frac)
        coeff = res[0]
    else:
        res = np.linalg.lstsq(A.T, b, rcond=1-svd_frac)
        coeff = res[0]

    # if nsvd is an array, returns array of coefficients
    return coeff

def mkpsf_loci(bg_opt, sci_opt, bg_sub, nsvd=None, svd_frac=0.9, **kwargs):
    """
    Input an array of nsvd to test variety of PCAs.
    Use svd_frac to perform a fraction power cut-off,
    which automatically selects some number of PCAs.

    Parameters
    ----------
    bg_opt : array_like
        Optimization region of reference observations.
    sci_opt : array_like
        Optimization region of science observations (single vector size of bg_opt.shape[1]).
    bg_sub : array_like
        Subtraction region of reference observations.
    """

    import numpy.ma as ma

    coeff = fit_loci_coeff(bg_opt, sci_opt, nsvd=nsvd, svd_frac=svd_frac, **kwargs)

    # Deal with NaNs
    inan_bg = np.isnan(bg_sub)
    if np.any(inan_bg):
        use_masked = True
        bg_sub = ma.masked_invalid(bg_sub)
    else:
        use_masked = False

    # These are mathematically equivalent
    # res = np.array([np.matmul(cf.T, bg_sub) for cf in coeff])
    # res = np.matmul(coeff, bg_sub)

    # If multiple dimensions, then coeff is a matrix
    cf_vals = coeff if (nsvd is not None) and (len(nsvd) > 1) else coeff.T

    # For 2D arrays, np.dot and np.matmul are equivalent
    res = ma.dot(cf_vals, bg_sub).data if use_masked else np.matmul(cf_vals, bg_sub)

    # if np.isnan(res).any():
    #     print(res.shape, res.size, np.isnan(res).sum())

    return res

def klip_subtract(sci_opt, bg_opt, numbasis=None, svd_frac=0.9,
                  remove_mean=False):
    """ Perform KLIP subtracton a sub-region of the data

    Parameters
    ----------
    sci_opt : array_like
        Science data to be subtracted
    bg_opt : array_like
        Background data to be used for PSF subtraction
    numbasis : int, optional
        Number of KL basis vectors to use
    svd_frac : float, optional
        Fraction of total variance to use for SVD cut-off
    remove_mean : bool, optional
        Remove the mean of the data before performing KLIP
    """

    # Any NaNs in the data?
    inan_bg = np.isnan(bg_opt)
    inan_sci = np.isnan(sci_opt)

    isnan_bg = np.any(inan_bg)
    isnan_sci = np.any(inan_sci)

    if remove_mean:
        # Use nanmean to ignore NaNs
        fmn_bg = np.nanmean if isnan_bg else np.mean
        fmn_sci = np.nanmean if isnan_sci else np.mean

        # Subtract mean of each frame data
        bg_mean_sub = bg_opt - fmn_bg(bg_opt, axis=1)[:, None]
        sci_mean_sub = sci_opt - fmn_sci(sci_opt)

        # Replace NaNs with zeros?
        if isnan_bg:
            bg_mean_sub[inan_bg] = 0
        if isnan_sci:
            sci_mean_sub[inan_sci] = 0
    else:
        # Replace NaNs with zeros?
        bg_mean_sub = bg_opt if not isnan_bg else np.nan_to_num(bg_opt)
        sci_mean_sub = sci_opt if not isnan_sci else np.nan_to_num(sci_opt)
    
    kpsf = mkpsf_klip(bg_mean_sub, sci_mean_sub, nsvd=numbasis, svd_frac=svd_frac)
    res = sci_mean_sub - kpsf
    
    return res

def loci_subtract(sci_opt, bg_opt, bg_sub, sci_sub, 
                  numbasis=None, svd_frac=0.9, remove_mean=False, **kwargs):
    """ Perform LOCI subtracton a sub-region of the data
    
    Parameters
    ----------
    sci_opt : array_like
        Science data within optimization region
    bg_opt : array_like
        Background (PSF) data frames within optimization region
    bg_sub : array_like
        Background (PSF) data frames within subtraction region
    sci_sub : array_like
        Science data within subtraction region
    numbasis : int, optional
        Number of KL basis vectors to use
    svd_frac : float, optional
        Fraction of total variance to use for SVD cut-off
    remove_mean : bool, optional
        Remove the mean of the data before performing fits

    Keyword Args
    ------------
    use_old : bool, optional
        Use older manual method of computing coefficients
    use_QR : bool, optional
        Use QR decomposition to compute coefficients
    """
        
    if remove_mean:
        # Any NaNs in the data?
        isnan_bg = np.any(np.isnan(bg_opt))
        isnan_sci = np.any(np.isnan(sci_opt))
        # Use nanmean to ignore NaNs
        fmn_bg = np.nanmean if isnan_bg else np.mean
        fmn_sci = np.nanmean if isnan_sci else np.mean

        bg_opt_mean = fmn_bg(bg_opt, axis=1)[:, None]
        sci_opt_mean = fmn_sci(sci_opt)

        bg_mean_opt  = bg_opt  - bg_opt_mean
        bg_mean_sub  = bg_sub  - bg_opt_mean
        sci_mean_opt = sci_opt - sci_opt_mean
        sci_mean_sub = sci_sub - sci_opt_mean
    else:
        bg_mean_opt  = bg_opt  
        bg_mean_sub  = bg_sub  
        sci_mean_opt = sci_opt 
        sci_mean_sub = sci_sub   

    # Create a PSF at the science subtraction region
    lpsf = mkpsf_loci(bg_mean_opt, sci_mean_opt, bg_mean_sub, 
                      nsvd=numbasis, svd_frac=svd_frac, **kwargs)
    
    return sci_mean_sub - lpsf

#################################################################
# Functions for masking optimization and subtraction regions
#################################################################

def define_annuli_bounds(annuli, IWA, OWA, annuli_spacing='constant'):
    """
    Defines the annuli boundaries radially

    Parameters
    ==========
    annuli: int
        number of annuli
    IWA: int 
        inner working angle (pixels)
    OWA: int 
        outer working angle (pixels)
    annuli_spacing: str
        how to distribute the annuli radially. Currently three options. 
        'constant' : (equally spaced),
        'log'      : (logarithmical expansion with r)
        'linear'   : (linearly expansion with r)

    Returns
    =======
    array of 2-element tuples that specify the beginning and end radius of that annulus

    """
    #calculate the annuli ranges
    if annuli_spacing.lower() == "constant":
        dr = float(OWA - IWA) / annuli
        rad_bounds = [(dr * rad + IWA, dr * (rad + 1) + IWA) for rad in range(annuli)]
    elif annuli_spacing.lower() == "log":
        # calculate normalization of log scaling
        unnormalized_log_scaling = np.log(np.arange(annuli) + 1) + 1
        log_coeff = float(OWA - IWA)/np.sum(unnormalized_log_scaling)
        # construct the radial spacing
        rad_bounds = []
        for i in range(annuli):
            # lower bound is either mask or end of previous annulus
            if i == 0:
                lower_bound = IWA
            else:
                lower_bound = rad_bounds[-1][1]
            upper_bound = lower_bound + log_coeff * unnormalized_log_scaling[i]
            rad_bounds.append((lower_bound, upper_bound))
    elif annuli_spacing.lower() == "linear":
        # scale linaer scaling to OWA-IWA
        linear_coeff = float(OWA - IWA) / np.sum(np.arange(annuli) + 1)
        dr = linear_coeff * (np.arange(annuli) + 1)
        bins = np.append([IWA], np.cumsum(dr) + IWA)
        r1_arr = bins[0:-1]
        r2_arr = bins[1:]
        rad_bounds = [(r1,r2) for r1, r2 in zip(r1_arr, r2_arr)]
    else:
        raise ValueError("annuli_spacing currently only supports 'constant', 'log', or 'linear'")

    # check to make sure the annuli are all greater than 1 pixel
    min_width = np.min(np.diff(rad_bounds, axis=1))
    if min_width < 1:
        raise ValueError("Too many annuli, some annuli are less than 1 pixel")

    return rad_bounds

def define_theta_bounds(subsections, degrees=True):
    """ Split angles between [-180,+180] into equal sizes"""
    
    dphi = 2 * np.pi / subsections
    phi_bounds = np.array([[dphi * phi_i - np.pi, dphi * (phi_i + 1) - np.pi] for phi_i in range(subsections)])
    phi_bounds[-1][1] = np.pi

    if degrees:
        return np.rad2deg(phi_bounds)
    else:
        return phi_bounds

def get_sub_regions(im, fwhm_pix, IWA_nfwhm=1, OWA_nfwhm=None,
                    sub_ann_rad=1, sub_ann_width=1, 
                    annuli_spacing='constant', constant_theta=False):
    """ Create an image of subtraction regions

    Parameters
    ==========
    im : ndarray
        Image template to get size
    fwhm_pix : float
        Number of pixels equal to PSF FWHM
    IWA_nfwhm : float
        Inner working angle in units of PSF FWHM
    OWA_nfwhm : float
        Outer working angle in units of PSF FWHM
    sub_ann_rad : float
        Radial width of annulus (units of fwhm)
    sub_ann_width : float
        Arc length of annulus (units of fwhm)
    annuli_spacing: str
        how to distribute the annuli radially. Currently three options. 
        'constant' : (equally spaced),
        'log'      : (logarithmical expansion with r)
        'linear'   : (linearly expansion with r)
    constant_theta : bool
        Do we want constant angle for each ring?
        Otherwise, the arc length (number of pixels) stays the same.
    """
    
    rho_im, th_im = dist_image(im, return_theta=True)
    
    # Determine outer radius of mask
    ny, nx = np.shape(im)
    OWA_nfwhm_max = np.sqrt((nx/2)**2 + (ny/2)**2) / fwhm_pix
    if (OWA_nfwhm is None) or (OWA_nfwhm > OWA_nfwhm_max):
        OWA_nfwhm = OWA_nfwhm_max
        
    assert IWA_nfwhm<OWA_nfwhm, "IWA_nfwhm must be less than OWA_nfwhm"
    
    # How many annuli to generate?
    n_annuli = round_int((OWA_nfwhm - IWA_nfwhm) / sub_ann_rad  + 0.5)
    dr_pix = sub_ann_rad * fwhm_pix
    
    # Generate list of radial bounds
    IWA_pix = IWA_nfwhm * fwhm_pix
    OWA_pix = IWA_pix + n_annuli*dr_pix
    rad_bounds = define_annuli_bounds(n_annuli, IWA_pix, OWA_pix, 
                                      annuli_spacing=annuli_spacing)
    
    # Either constant theta or constant pixel size
    if constant_theta:
        # Base number of theta sections on first radius and nfwhm_ann_width
        pix_dist = np.mean(rad_bounds[0])
        circ_npix = 2*np.pi*pix_dist
        nsect = int(circ_npix / (sub_ann_width*fwhm_pix))
        
        th_bounds = define_theta_bounds(nsect, degrees=True)
        # Repeat for each annulus
        th_bounds_all = np.tile(th_bounds, n_annuli).reshape([-1,n_annuli,2]).transpose([1,0,2])
    else:
        th_bounds_all = []
        for rvals in rad_bounds:
            pix_dist = np.mean(rvals)
            circ_npix = 2*np.pi*pix_dist
            nsect = int(circ_npix / (sub_ann_width*fwhm_pix))
            
            th_bounds = define_theta_bounds(nsect, degrees=True)
            th_bounds_all.append(th_bounds)
    
    # Create region mask
    region_mask = np.zeros_like(rho_im).astype('int')
    ii = 1
    for jj in range(n_annuli):
        r1, r2 = rad_bounds[jj]
        # print(r1,r2,r2-r1)
        th_bounds = th_bounds_all[jj]
        for th1, th2 in th_bounds:
            ind = (rho_im>=r1) & (rho_im<=r2) & (th_im>=th1) & (th_im<=th2)
            region_mask[ind] = ii
            ii += 1

    # Ensure consecutive integers
    region_mask_fin = np.zeros_like(region_mask)
    iuniq = np.unique(region_mask)
    # Exclude 0s
    iuniq = iuniq[iuniq>0]
    for ii, iu in enumerate(iuniq):
        region_mask_fin[region_mask==iu] = ii+1

    return region_mask_fin
        
def get_opt_regions(region_mask, fwhm_pix, opt_ann_rad=None, 
                    opt_ann_width=0.75, nfwhm_sep=0, exclude_sub=True):
    """ Create array of optimization region masks for each subtraction region

    For LOKI style subtractions, we utilize optimization regions that are
    adjacent to the subtraction region. This function creates a series of
    image masks definining those optimization regions for each subtraction
    region.

    Parameters
    ==========
    region_mask : image
        Input of regions generated with `get_sub_regions`.
        Values set to 0 will be ignored.
    fwhm_pix : float
        Number of pixels equal to PSF FWHM
    opt_ann_rad : float
        Radial width of annulus (units of fwhm)
        If set to None, then will match region_mask thicknesses
    opt_ann_width : float
        Arc length of annulus (units of fwhm)
    nfwhm_sep : float
        Separation between subtraction and optimization regions (units of fwhm)
    exclude_sub : bool
        Do we want to exclude the subtraction region from the optimization
        region?
    """
    
    rho_im, th_im = dist_image(region_mask, return_theta=True)
    # Used for indexing thetas <-180 or >+180
    th_im_m180 = th_im.copy()
    th_im_m180[th_im>0] = th_im[th_im>0] - 360
    th_im_p180 = th_im.copy()
    th_im_p180[th_im<0] = th_im[th_im<0] + 360
        
    mask_vals = np.unique(region_mask)
    # Exclude 0
    mask_vals = mask_vals[mask_vals>0]
    masks_opt = []
    for val in mask_vals:
        ind_sub = (region_mask == val)
        # Get center location
        rho_mn = 0.5 * (rho_im[ind_sub].max() + rho_im[ind_sub].min())
        
        # Create annulus mask
        if opt_ann_rad is None:
            rho_min, rho_max = rho_im[ind_sub].min(), rho_im[ind_sub].max()
        else:
            rho_min, rho_max = rho_mn + np.array([-1,1])*opt_ann_rad*fwhm_pix/2
        annulus_mask = (rho_im >= rho_min) & (rho_im < rho_max)

        # th_im goes from -180 to +180
        # Mask regions do not span that the division
        th_ind = th_im[ind_sub]
        
        # Record edges of subtraction region
        th_lims = np.array([th_ind.min(), th_ind.max()])
        th_mn = np.mean(th_lims)
        
        # Get regions on either side of subtraction region
        # These are the start locations
        offset_start = ((nfwhm_sep*fwhm_pix) / (2*np.pi*rho_mn)) * 360
        offset_end   = ((opt_ann_width*fwhm_pix) / (2*np.pi*rho_mn)) * 360
        # Start locations for two sides of symmetric optimization region
        th_opt_start = th_mn + np.array([-1,1]) * offset_start
        # Minor offsets to make sure opt widths go beyond subtraction region
        th_opt_start2 = th_lims + np.array([-1,1]) * offset_start
        th_opt_end    = th_opt_start2 + np.array([-1,1]) * offset_end
        
        th_ims = [th_im, th_im_m180, th_im_p180]

        # First region
        th1, th2 = th_opt_start[0], th_opt_end[0]
        if th1<=-180 or th2<=-180:
            th_ind = 1
        elif th1>=180 or th2>=180:
            th_ind = 2
        else:
            th_ind = 0

        theta_mask1 = (th_ims[th_ind]<=th1) & (th_ims[th_ind]>=th2)

        # Second region
        th1, th2 = th_opt_start[1], th_opt_end[1]
        if th1<=-180 or th2<=-180:
            th_ind = 1
        elif th1>=180 or th2>=180:
            th_ind = 2
        else:
            th_ind = 0

        theta_mask2 = (th_ims[th_ind]>=th1) & (th_ims[th_ind]<=th2)
        
        # Final optimization region mask
        ind_opt = (theta_mask1 | theta_mask2) & annulus_mask

        # Exclude anything in original ind_sub?
        if exclude_sub:
            ind_opt = ind_opt & ~ind_sub
        
        masks_opt.append(ind_opt)
    
    masks_opt = np.array(masks_opt)
    
    return masks_opt

def create_region_masks(shape, fwhm_pix, kw_sub={}, kw_opt={}):

    # Subtraction and optimization region masks
    kwargs_sub = {
        'IWA_nfwhm' : kw_sub.get('IWA_nfwhm', 0.0),
        'OWA_nfwhm' : kw_sub.get('OWA_nfwhm', None),
        'sub_ann_rad'   : kw_sub.get('sub_ann_rad', 2.0),
        'sub_ann_width' : kw_sub.get('sub_ann_width', 2.0),
        'annuli_spacing' : kw_sub.get('annuli_spacing', 'constant'),
        'constant_theta' : kw_sub.get('constant_theta', False),
    }
    kwargs_opt = {
        'opt_ann_rad'   : kw_opt.get('opt_ann_rad'),
        'opt_ann_width' : kw_opt.get('opt_ann_width', 1),
        'nfwhm_sep'   : kw_opt.get('nfwhm_sep', 0),
        'exclude_sub' : kw_opt.get('exclude_sub', True),
    }
    if kwargs_opt['opt_ann_rad'] is None:
        kwargs_opt['opt_ann_rad'] = kwargs_sub['sub_ann_rad']+1

    # Create subtraction and optimization regions (same)
    sub_masks = get_sub_regions(np.zeros(shape), fwhm_pix, **kwargs_sub)
    opt_masks = get_opt_regions(sub_masks, fwhm_pix, **kwargs_opt)

    return sub_masks, opt_masks


def plot_masks(sub_masks, opt_masks=None, omask_index=None, pixscale=None,
               save_name=None, save_dir=None, return_figax=False, 
               verbose=True, **kwargs):
    """Plot the subtraction and optimization masks

    Parameters
    ----------
    sub_masks : numpy.ndarray
        Image array of the subtraction regions
    opt_masks : numpy.ndarray, optional
        Image array of the optimization regions. If set to None,
        then the optimization regions will not be plotted.
    omask_index : int, optional
        Index of the optimization region to plot. By default, will
        choose the middle region.
    pixscale : float, optional
        Pixel scale of the image. If set to None, then the axes
        will be in units of pixels.
    save_name : str, optional
        Name of the file to save the plot. Default is None
        means that the plot will not be saved.
    save_dir : str, optional
        Directory to save the plot. Default is None, which
        means that the current working directory will be used.
    return_figax : bool, optional
        If True, return the figure and axes objects. Default is False.
    verbose : bool, optional
        If True, print out information about the saved file. Default is True.
    """

    if opt_masks is None:
        fig, ax = plt.subplots(1,1,figsize=(5,5))
        axes = [ax]
    else:
        fig, axes = plt.subplots(1,2,figsize=(10,5))

    if verbose:
        nmasks = len(np.unique(sub_masks[sub_masks>0]))
        print(f"Number of masks: {nmasks}")

    ny, nx = sub_masks.shape

    extent = np.array([-nx,nx,-ny,ny]) / 2
    if pixscale is not None:
        extent *= pixscale
        xlabel = ylabel = 'arcsec'
    else:
        xlabel = ylabel = 'pixels'

    axes[0].imshow(sub_masks, extent=extent)
    axes[0].set_title('Subtraction Regions')

    # Are we also plotting optimization regions?
    if opt_masks is not None:
        omask_index = int(len(opt_masks)/2) if omask_index is None else omask_index
        axes[1].imshow(opt_masks[omask_index], extent=extent)
        axes[1].set_title('Optimization Region Example')

    for ax in axes:
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_aspect('equal')

        ax.tick_params(axis='both', color='w', which='both')
        for k in ax.spines.keys():
            ax.spines[k].set_color('w')

    fig.tight_layout()

    if save_name is not None:
        if save_dir is None:
            save_dir = os.getcwd()
        save_path = os.path.join(save_dir, save_name)
        if verbose:
            print(f"Saving: {save_path}")
        fig.savefig(save_path, bbox_inches='tight', pad_inches=0)

    if return_figax:
        return fig, axes


def plot_mask_images(subsize, fwhm_pix, bin=1, do_loci=True,
                     kwargs_sub={}, kwargs_opt={}, **kwargs):
    """ Plot subtraction and optimization regions
    
    Parameters
    ----------
    pca_params : dict

    Keyword Args
    ------------
    omask_index : int, optional
        Index of the optimization region to plot. By default, will
        choose the middle region.
    pixscale : float, optional
        Pixel scale of the image. If set to None, then the axes
        will be in units of pixels.
    save_name : str, optional
        Name of the file to save the plot. Default is None
        means that the plot will not be saved.
    save_dir : str, optional
        Directory to save the plot. Default is None, which
        means that the current working directory will be used.
    return_figax : bool, optional
        If True, return the figure and axes objects. Default is False.
    verbose : bool, optional
        If True, print out information about the saved file. Default is True.
    """

    # Create subtraction and optimization regions (same)
    nx = ny = int(subsize / bin)
    im = np.zeros((ny,nx))
    sub_masks = get_sub_regions(im, fwhm_pix, **kwargs_sub)

    # Only produce optimization regions if LOCI
    plot_opt = do_loci
    if plot_opt:
        opt_masks = get_opt_regions(sub_masks, fwhm_pix, **kwargs_opt)
    else:
        opt_masks = None

    # Plot masks
    return plot_masks(sub_masks, opt_masks=opt_masks, **kwargs)


def corr_all(image, psf_arr, bp_image=None, bp_psf=None, sub_size=32, inner_rad=0):
    """ Get array of corr coeff for image to PSFs """

    from webbpsf_ext.image_manip import crop_image
    from webbpsf_ext.imreg_tools import correl_images

    # Create mask to exclude central region
    if inner_rad > 0:
        rho = dist_image(np.zeros((sub_size, sub_size)))
        mask_good = rho > inner_rad
    else:
        mask_good = np.ones((sub_size, sub_size), dtype=bool)

    if bp_image is not None:
        mask_good = mask_good & ~bp_image
    if bp_psf is not None:
        if len(bp_psf.shape) == 3:
            bp_psf = np.logical_and.reduce(bp_psf, axis=0)
        mask_good = mask_good & ~bp_psf

    # Cropped versions
    imsub = crop_image(image, sub_size)
    psfsub = crop_image(psf_arr, sub_size)
    # Get correlation matrix of each original PSF to each other
    return correl_images(imsub, psfsub, mask=mask_good)


def pca_subtraction(imarr, psfarr, fwhm_pix, do_loci=True, 
                    npsf_max=None, numbasis=None, svd_frac=0.9, remove_mean=False,
                    more_progress=False, kw_sub={}, kw_opt={}, **kwargs):
    """ Perform PCA subtraction on a set of images
    
    Parameters
    ----------
    imarr : array_like
        Image cube to be subtracted
    psfarr : array_like
        PSF library to be used for subtraction
    fwhm_pix : float
        Number of pixels equal to PSF FWHM
    do_loci : bool
        Use LOCI subtraction instead of KLIP
    npsf_max : int, optional
        Maximum number of PSFs to use in library to generate PCA basis
    numbasis : int or array_like, optional
        Number of PCA basis vectors to use for generating final PSF. 
        If set to None, then the number of singular values is chosen to be 
        the minimum number that account for svd_frac of the variance in the data. 
        If set to an integer, then will use that number of basis vectors. 
        If set to an array, then will loop through each value.
    svd_frac : float, optional
        Fraction of total variance to use for SVD cut-off
    remove_mean : bool, optional
        Remove the mean of the data and PSFs prior to generating PCA basis
        and performing fits.
    more_progress : bool, optional
        Show more detailed progress bar
    kw_sub : dict, optional
        Keyword arguments for creating subtraction region masks
    kw_opt : dict, optional
        Keyword arguments for creating optimization region masks

    """

    shape_orig = imarr.shape
    ndim = len(shape_orig)
    if ndim==3:
        nint, ny, nx = shape_orig
    elif ndim==2:
        nint, ny, nx = (1, shape_orig[0], shape_orig[1])
        imarr = imarr.reshape([nint,ny,nx])

    psf_shape = psfarr.shape
    npsf = psf_shape[0]
    if len(psf_shape) != 3 or npsf==1:
        raise ValueError(f"PSF array must be image cube and have multiple frames {psf_shape}")
    
    do_klip = not do_loci

    # Create subtraction and optimization region masks
    sub_masks, opt_masks = create_region_masks(imarr[0].shape, fwhm_pix, kw_sub=kw_sub, kw_opt=kw_opt)
    if do_klip:
        opt_masks = None
    # Number of sub region masks
    nmasks = len(np.unique(sub_masks[sub_masks>0]))

    # print(numbasis, npsf_max, npsf)

    # Maximum number of PSFs to use in library
    if (npsf_max is None) or (npsf_max > npsf):
        npsf_max = npsf
    if numbasis is None:
        nbasis_sets = 1
    elif isinstance(numbasis, (tuple, list, np.ndarray)):
        numbasis = np.asarray(numbasis)
        # Filter out numbasis where greater than npsf_max
        ind_out_of_range = numbasis > npsf_max
        if np.any(ind_out_of_range):
            _log.warning(f"numbasis values {numbasis[ind_out_of_range]} are greater than npsf_max={npsf_max}")
        numbasis = numbasis[~ind_out_of_range]
        nbasis_sets = len(numbasis)
        if nbasis_sets == 0:
            raise ValueError("All numbasis values are greater than npsf_max")
    elif isinstance(numbasis, int):
        if numbasis > npsf_max:
            raise ValueError(f"numbasis={numbasis} is greater than npsf_max={npsf_max}")
        numbasis = np.array([numbasis])
        nbasis_sets = 1

    # Initialize array of de-rotated and subtracted images
    fin_arr = np.zeros([nint, nbasis_sets, ny, nx])

    # print(fin_arr.shape)

    # Loop through all images
    for j in trange(nint, desc=f"Images", leave=False):
        im = imarr[j]

        # Get top correlated PSFs
        corr_vals = corr_all(im, psfarr, sub_size=32, inner_rad=5)
        imax = np.min([npsf_max, len(corr_vals)])
        isort = np.argsort(corr_vals)[::-1][0:imax]
        psf_lib = psfarr[isort]
        # corr_im = corr_vals[isort]

        psf_mean = np.nanmean(psf_lib, axis=0)
        psf_lib_resid = psf_lib - psf_mean
        im_resid = im - psf_mean

        # Loop through all subtraction regions
        sub_res = np.zeros([nbasis_sets,ny,nx])
        jj_arr = np.arange(1, nmasks+1)
        jj_iter = tqdm(jj_arr, leave=False, desc="Subtraction Region") if more_progress else jj_arr
        for jj in jj_iter:
            # Select optimization and subtraction regions
            # For KLIP, these are the same region
            ind_sub = (sub_masks == jj)
            ind_opt = ind_sub if do_klip else opt_masks[jj-1] 

            sci_sub = im_resid[ind_sub]  # Science subtraction region
            bg_opt = psf_lib_resid[:,ind_opt] # PSF optimization region

            # KLIP or LOCI PSF subtraction
            if do_loci:
                sci_opt = im_resid[ind_opt]        # Science optimization region
                bg_sub = psf_lib_resid[:,ind_sub]  # PSF subtraction region
                sub_res[:,ind_sub] = loci_subtract(sci_opt, bg_opt, bg_sub, sci_sub, 
                                                   numbasis=numbasis,
                                                   svd_frac=svd_frac,
                                                   remove_mean=remove_mean,
                                                   **kwargs)
            else:
                sub_res[:,ind_sub] = klip_subtract(sci_sub, bg_opt, 
                                                   numbasis=numbasis, 
                                                   svd_frac=svd_frac,
                                                   remove_mean=remove_mean)

        fin_arr[j] = sub_res

    return fin_arr.squeeze()


def run_pca_subtraction(imarr, psfarr, pca_params):
    """ Run PCA subtraction on a set of parameters """

    # Unpack parameters
    fwhm_pix = pca_params['fwhm_pix']
    do_loci = pca_params['loci']
    npsf_max = pca_params['numpsfs_max']
    numbasis = pca_params['numbasis']
    svd_frac = pca_params['svd_frac']
    kw_sub = pca_params['kwargs_sub']
    kw_opt = pca_params['kwargs_opt']

    return pca_subtraction(imarr, psfarr, fwhm_pix, do_loci=do_loci, 
                           npsf_max=npsf_max, numbasis=numbasis, 
                           svd_frac=svd_frac, kw_sub=kw_sub, kw_opt=kw_opt)

# def run_pca_subtraction(imarr, psfarr, fwhm_pix, do_loci=True, 
#                         npsf_max=None, numbasis=None, svd_frac=0.9, remove_mean=False,
#                         more_progress=True, kw_sub={}, kw_opt={}, **kwargs):


def build_pca_dict(fwhm_pix, **kwargs):
    """ Build PCA parameter dictionary

    Returns a dictionary of settings to pass to various analysis functions.
    There are multiple keywords that will be set by default and can be modified 
    by passing via keyword arguments. 
    
    Example output:

    pca_params = {
        'fwhm_pix' : fwhm_pix, # Number of pixels equal to PSF FWHM
        'loci'        : True,  # Use LOCI instead of KLIP?
        'remove_mean' : False,  # Remove mean from each frame before PCA?
        'numpsfs_max' : 100,    # Maximum number of PSFs to use for decomposition
        'numbasis'    : [1,5,10,50],  # Number of KL basis vectors to use
        'svd_frac'    : 0.9, # Fraction of variance to describe with KL basis vectors

        'kwargs_sub' : {}, # Subtraction regions config
        'kwargs_opt' : {}, # Optimization regions config
    }


    # Subtraction regions config
    kwargs_sub = {
        'IWA_nfwhm' : 0,     # Inner working angle in units of FWHM
        'OWA_nfwhm' : None,  # Outer working angle in units of FWHM
        'sub_ann_rad'   : 3, # Radial width of annulus (units of FWHM)
        'sub_ann_width' : 3, # Arc length of annulus (units of FWHM)
        'annuli_spacing' : 'constant', # Constant, log, or linear
        'constant_theta' : False,      # Do we want constant angle for each ring?
    }    

    # Optimization regions config (only for LOCI)
    kwargs_opt = {
        'opt_ann_rad'   : None, # Radial width of annulus (units of FWHM)
        'opt_ann_width' : 4,    # Arc length of annulus (units of FWHM)
        'nfwhm_sep'   : 0,      # Separation between subtraction and optimization regions (units of FWHM)
        'exclude_sub' : True,   # Exclude the subtraction region from the optimization region?
    }

    Parameters
    ----------
    param_dict : dict
        Dictionary of parameters from the main reduction script.
    """


    # PCA Parameters
    pca_params = {
        'fwhm_pix' : fwhm_pix, # Number of pixels equal to PSF FWHM

        # 'bin' : 1,               # Image binning factor
        # 'navg_ints_sci' : None,  # Number of sci INTS to average together
        # 'navg_ints_ref' : None,  # Number of ref INTS to average together
        # 'scale_psfs'    : False, # Scale PSFs by flux before subtracting and combining?
        # 'sat_rad_asec'  : None,  # Saturation radius in arcsec

        # 'add_fake' : False,    # Add Fake Planets?

        'loci'        : True,  # Use LOCI instead of KLIP?
        'remove_mean' : False,  # Remove mean from each frame before PCA?
        'numpsfs_max' : 100,    # Maximum number of PSFs to use for decomposition
        'numbasis'    : [1,5,10,50],  # Number of KL basis vectors to use
        'svd_frac'    : 0.9, # Fraction of variance to describe with KL basis vectors
    }

    # Subtraction regions config
    kwargs_sub = {
        'IWA_nfwhm' : 0,
        'OWA_nfwhm' : None,
        'sub_ann_rad'   : 3, # In terms of FWHM
        'sub_ann_width' : 3, # In terms of FWHM
        'annuli_spacing' : 'constant',
        'constant_theta' : False,
    }
    for k in kwargs_sub.keys():
        kwargs_sub[k] = kwargs.pop(k, kwargs_sub[k])

    # Optimization regions config
    kwargs_opt = {
        'opt_ann_rad'   : None,
        'opt_ann_width' : 2,
        'nfwhm_sep'   : 0,
        'exclude_sub' : True,
    }
    for k in kwargs_opt.keys():
        kwargs_opt[k] = kwargs.pop(k, kwargs_opt[k])
    # Default the radius of the optimization region to be 1 FWHM larger than the subtraction region
    if kwargs_opt['opt_ann_rad'] is None:
        kwargs_opt['opt_ann_rad'] = kwargs_sub['sub_ann_rad']+1

    # Save to dictionary
    pca_params['kwargs_sub'] = kwargs_sub
    pca_params['kwargs_opt'] = kwargs_opt

    # Update defaults with any user inputs
    pca_params.update(**kwargs)

    return pca_params



# From https://github.com/joolof/DDiT as a comparison to VIP's scattered light module
# This code is not used in the current version of pynrc, but is kept here for reference.

import sys
import time
import numpy as np
import matplotlib.pyplot as plt
# -----------------------------------------------------------
np.seterr(over='ignore')
# -----------------------------------------------------------------------------
# Define some constants
# -----------------------------------------------------------------------------
CC = 2.9979e+14     # in microns/s 
HH = 6.6262e-27
KK = 1.3807e-16
# -----------------------------------------------------------
class Disk(object):
    """
    Class to compute images of debris disks, in polarized light and total intensity.
    """
    def __init__(self, nx = 300, pixscale = 0.01226, nframe = 150, thermal = False, dpc = None, gaussian = False, nm = 2., engler = False, threshold = 1.e-2):
        """
        Class to compute synthetic images of debris disks. To compute a model, simply do the following:
        > disk = Disk()
        > disk.compute_model(a = 0.8, e = 0.1, incl = 60.)

        And the images will be stored in disk.intensity and disk.polarized. More details in the README.md file.
        """

        """
        Some checks on the input parameters
        """
        if type(nx) is not int: self._error_msg('The parameter \'nx\' should be an integer ')
        if type(nframe) is not int: self._error_msg('The parameter \'nframe\' should be an integer ')
        if nframe%2 == 1:
            print('It is probably better if \'nframe\' is an even number.')
        if ((thermal) and (dpc is None)):
            self._error_msg('To compute thermal images, you need to provide a distance \'dpc\' in pc. This is because of the way the T(r) is defined.')
        self._threshold = threshold
        self._nm = nm
        self._engler = engler
        self._nx = nx
        if nx%2 ==0:
            self._cx = self._nx//2 - 0.5
        else:
            self._cx = self._nx//2
        self._nframe = nframe
        self._pixscale = pixscale
        self._thermal = thermal
        self._gaussian = gaussian
        self._dpc = dpc
        self._xlim = self._cx * self._pixscale
        self._Xin, self._Yin = (np.mgrid[0:self._nx, 0:self._nx] - self._cx) * self._pixscale
        self._xm, self._ym = np.zeros(shape=(self._nx, self._nx)), np.zeros(shape=(self._nx, self._nx))
        """
        Define some variables that will contain the:
            - total intensity image
            - polarized intensity image
            - scattering angle at the midplane
            - distance to the star at the midplane
            - azimuth angle at the midplane
        """
        self.polarized = np.zeros(shape=(self._nx, self._nx))
        self.intensity = np.zeros(shape=(self._nx, self._nx))
        self.scattering = np.zeros(shape=(self._nx, self._nx))
        self.distance = np.zeros(shape=(self._nx, self._nx))
        self.azimuth = np.zeros(shape=(self._nx, self._nx))
        """
        Some trigonometry variables
        """
        self._cs, self._ss, self._st, self._st2, self._to = 0., 0., 0., 0., 0.
        """
        The geometric parameters for the disk, with some default values.
        """
        self._a, self._incl, self._pa, self._pin, self._pout, self._e, self._omega, self._opang, self._dr, self._pmid, self._da, self._gamma = 1., 0.1, 152.1*np.pi/180., 25., -2.5, 0., 0.,0.04, 0.05, None, None, 2.0
        """
        Parameters for the phase functions
        """
        self._gsca, self._gpol = 0., 0.
        self.theta, self.s11, self.s12 = None, None, None
        """
        Parameters for the thermal images
        """
        self._nu = None

    @property
    def a(self):
        return self._a

    @a.setter
    def a(self, a):
        self._a = a

    @property
    def da(self):
        return self._da

    @da.setter
    def da(self, da):
        self._da = da

    @property
    def e(self):
        return self._e

    @e.setter
    def e(self, e):
        self._e = e

    @property
    def pin(self):
        return self._pin

    @pin.setter
    def pin(self, pin):
        self._pin = pin

    @property
    def pmid(self):
        return self._pmid

    @pmid.setter
    def pmid(self, pmid):
        self._pmid = pmid

    @property
    def opang(self):
        return self._opang

    @opang.setter
    def opang(self, opang):
        self._opang = opang

    @property
    def gamma(self):
        return self._gamma

    @gamma.setter
    def gamma(self, gamma):
        self._gamma = gamma

    @property
    def gpol(self):
        return self._gpol

    @gpol.setter
    def gpol(self, gpol):
        self._gpol = gpol

    @property
    def gsca(self):
        return self._gsca

    @gsca.setter
    def gsca(self, gsca):
        self._gsca = gsca

    @property
    def dr(self):
        return self._dr

    @dr.setter
    def dr(self, dr):
        self._dr = dr

    @property
    def pout(self):
        return self._pout

    @pout.setter
    def pout(self, pout):
        self._pout = pout

    @property
    def incl(self):
        return self._incl * 180. / np.pi

    @incl.setter
    def incl(self, incl):
        self._incl = incl * np.pi/180.

    @property
    def pa(self):
        return -self._pa * 180. / np.pi

    @pa.setter
    def pa(self, pa):
        self._pa = -pa * np.pi/180.

    @property
    def omega(self):
        return self._omega * 180. / np.pi

    @omega.setter
    def omega(self, omega):
        self._omega = omega * np.pi/180.

    """
    Get the entry and exit points of an ellipse in the [y,z] plane
    """
    def _get_sphere(self):
        """
        Intersection points with an ellipse along the [y,z] plane
        """
        ye, ys = np.zeros(shape=(self._nx, self._nx)), np.zeros(shape=(self._nx, self._nx))
        ze, zs = np.zeros(shape=(self._nx, self._nx)), np.zeros(shape=(self._nx, self._nx))
        xe = np.zeros(shape=(self._nx, self._nx))
        """
        rmax is the maximum radius that should contain most of the density of the disk.
        It should be a * (1 - e**2) / (1-e) which simplyfies as a*(1+e)
        """
        rmax = self._threshold**(1.e0 / self.pout) * self._a * (1.e0 + self._e)
        radius = rmax**2. - self._xm**2.
        radius[(radius<0)] = 0.
        radius = np.sqrt(radius)
        height = 3. * radius * self._to
        """
        Here I am solving the intersections of two equations:
          + (y/radius)**2. + (z/height)**2. = 1.
          + z = (y - ym) / tan(inclination)
        The first one is an ellipse and the second one is the line of sight intercepting the midplane at ym.
        Replacing z in the first equation with the expression from the second equation yields a second
        degree equation.
        """
        delta = self._ym**2. * self._st2**2. * radius ** 4. - (height**2. + self._st2 * radius**2.) * (self._st2 * self._ym**2. * radius**2. - height**2. * radius**2.)
        sel = (delta > 0.)
        ye[sel] = ((self._ym[sel] * self._st2 * radius[sel]**2.) + np.sqrt(delta[sel])) / (height[sel]**2. + self._st2 * radius[sel]**2.)
        ys[sel] = ((self._ym[sel] * self._st2 * radius[sel]**2.) - np.sqrt(delta[sel])) / (height[sel]**2. + self._st2 * radius[sel]**2.)
        ze[sel] = self._st * (ye[sel] - self._ym[sel])
        zs[sel] = self._st * (ys[sel] - self._ym[sel])
        xe[sel] = self._xm[sel]

        return xe, ye, ze, ys, zs

    """
    Compute the model for the total and polarized intensity.
    """
    def compute_model(self, **kwargs):
        """
        Method to compute a synthetic image of a debris disk

        INPUTS:
          - a: reference semi-major axis in arcseconds
          - incl: inclination in degrees
          - pa: position angle in degrees
          - pin: inner slope
          - pout: outer slope
          - gsca: HG coefficient for total intensity
          - gpol: HG coefficient for polarized intensity
          - e: eccentricity
          - omega: argument of pericenter in degrees
          - opang: opening angle of the disk
          - s11: the scattered light phase function
          - s12: the polarized light phase function
        """
        """
        Check the input parameters
        And put the angles in radians
        """
        self._check_parameters(kwargs)
        """
        Compute some cos, sin, and tan
        """
        self._trigonometry()
        """
        Need to re-initialize the arrays to zero, in case I compute two models one after the other
        """
        self.polarized = np.zeros(shape=(self._nx, self._nx))
        self.intensity = np.zeros(shape=(self._nx, self._nx))
        """
        Project the disk according to the inclination and position angle
        """
        self._xm = ((np.cos(self._pa) * self._Xin + np.sin(self._pa) * self._Yin))
        self._ym = ((np.sin(self._pa) * self._Xin - np.cos(self._pa) * self._Yin)) / self._cs
        self.distance = np.sqrt(self._xm**2. + self._ym**2.)
        if self._nx % 2 == 1:
            self.distance[self._cx, self._cx] = 1. # This is to avoid a division by zero when computing the scattering angle in the midplane.
        """
        Compute the azimuth and scattering angles in the disk midplane, so
        that they can be used outside of the class.

        The azimuth angle should be zero along the projected minor axis of the disk.
        """
        self.azimuth = (np.arctan2(self._ym,self._xm) + np.pi/2.) % (2. * np.pi) - np.pi
        self.scattering = np.arccos((self._ss * self._ym)/self.distance)
        """
        Get the entry and exit points for the upper and lower layers
        """
        xe, ye, ze, ys, zs = self._get_sphere()
        """
        Compute the final image
        """
        self._get_flux(xe, ye, ze, ys, zs)

    """
    Method to compute the emission between the entry and exit points.
    """
    def _get_flux(self, xe, ye, ze, ys, zs):
        """
        The volume of each cells
        The line of sight is divided as follows:
        for i in range(nframe):
            zi = (ze-zs) * i / (nframe - 1) + zs
        So then I can compute the lenght of a given cell which is (ze-zs)/(nframe -1)
        And the volume of one cell is sqrt(Delta_z**2. + Delta_y**2) * pixelscale**2.
        """
        volume = np.sqrt(((ze-zs)**2. + (ye-ys)**2.))/(self._nframe-1.)*self._pixscale**2.
        for i in range(self._nframe-1):
            """
            Define some variables, that need to be re-initialized for every iteration
            This slows down the code a bit, but it is more accurate and prevents some unfortunate
            NaNs in some cases (especially if the position angle is 0 degrees for instance.

            I had tried to define them once, and set them to 0., but that did not improve the time.
            For clarity, I leave it as it is now.
            """
            densr = np.zeros(shape=(self._nx, self._nx))
            densz = np.zeros(shape=(self._nx, self._nx))
            """
            Get the middle points of each cells.
            """
            zi = zs + (ze - zs) * (i+0.5) / (self._nframe-1)
            yi = ys + (ye - ys) * (i+0.5) / (self._nframe-1)
            """
            Compute the 2D and 3D distances, with their respective
            selection where they are different than 0
            """
            dist2d = np.sqrt(xe**2. + yi**2.)
            sel2d = (dist2d > 0.)
            dist3d = np.sqrt(xe**2. + yi**2. + zi**2.)
            sel3d = (dist3d > 0.)
            """
            Define the radial and vertical density structure.
            For the radial one, I need the azimuthal angle at the midplane, which is np.arctan2(yi, xe)
            """
            r_ref = self._a * (1.e0 - self._e * self._e) / (1.e0 + self._e * np.cos(np.arctan2(yi, xe) + self._omega))
            if self._gaussian:
                densr[sel2d] = np.exp(-(r_ref[sel2d] - dist2d[sel2d])**2. / (2 * self._dr**2.))
            else:
                if self._pmid is None:
                    densr[sel2d] = ((dist2d[sel2d]/r_ref[sel2d])**(- self._nm * self._pout) + (dist2d[sel2d]/r_ref[sel2d])**(- self._nm * self._pin))**(-1./self._nm)
                else:
                    r_ref2 = (self._a + self._da) * (1.e0 - self._e * self._e) / (1.e0 + self._e * np.cos(np.arctan2(yi, xe) + self._omega))
                    dens2 = np.zeros(shape=(self._nx, self._nx))
                    dens3 = np.zeros(shape=(self._nx, self._nx))
                    dens2[sel2d] = ((dist2d[sel2d]/r_ref[sel2d])**(-self._nm*self._pin) + (dist2d[sel2d]/r_ref[sel2d])**(-self._nm * self._pmid))**(-1./self._nm)
                    dens3[sel2d] = dens2[sel2d] * (dist2d[sel2d]/r_ref2[sel2d])**(self._pout-self._pmid)
                    densr[sel2d] = (dens2[sel2d]**(-self._nm) + dens3[sel2d]**(-self._nm))**(-1./self._nm)
                    densr[sel2d] = densr[sel2d] / np.max(densr[sel2d])
            if self._engler:
                densz[sel2d] = np.exp(-np.log(2) * zi[sel2d]**2 / ((self._to * dist2d[sel2d])**2.))
            else:
                densz[sel2d] = np.exp(-(np.abs(zi[sel2d]) / (self._to * dist2d[sel2d]))**self._gamma)
            # densz[sel2d] = np.exp(-zi[sel2d]**2 / (2. * (self._to * dist2d[sel2d])**2.))
            if self._thermal:
                temperature = np.zeros(shape=(self._nx, self._nx))
                expterm = np.zeros(shape=(self._nx, self._nx))
                bplanck = np.zeros(shape=(self._nx, self._nx))
                """
                Following Eq. 3 of Wyatt 2008, without the stellar luminosity
                """
                temperature[sel3d] = 278.3 / np.sqrt(dist3d[sel3d] * self._dpc)
                expterm[sel3d] = np.exp(self._nu * HH / KK / temperature[sel3d])
                bplanck[sel3d] = 2. * HH * self._nu**3. / (expterm[sel3d]-1.0) / (CC * CC)
                self.intensity[sel3d] += densr[sel3d] * densz[sel3d] * volume[sel3d] * bplanck[sel3d]
            else:
                psca = np.zeros(shape=(self._nx, self._nx, 2))
                ppol = np.zeros(shape=(self._nx, self._nx, 2))
                theta = np.zeros(shape=(self._nx, self._nx))
                costheta = np.zeros(shape=(self._nx, self._nx))
                density = np.zeros(shape=(self._nx, self._nx))
                """
                Compute the cosine of the scattering angle
                I need to make a selection where dist3d != 0, to avoid division by 0 when computing it.
                """
                if self.theta is None:
                    costheta[sel3d] = (self._ss * yi[sel3d] - self._cs * zi[sel3d]) / dist3d[sel3d]
                    hg = (1.e0 - self._gsca**2.) / (4. * np.pi * (1.e0 + self._gsca**2. - 2. * self._gsca * costheta[sel3d])**(1.5))
                    phg = (1.e0 - self._gpol**2.) / (4. * np.pi * (1.e0 + self._gpol**2. - 2. * self._gpol * costheta[sel3d])**(1.5)) * (1.e0 - costheta[sel3d]**2.) / (1.e0 + costheta[sel3d]**2.)
                    psca[sel3d,0] = hg
                    psca[sel3d,1] = hg
                    ppol[sel3d,0] = phg
                    ppol[sel3d,1] = phg
                else:
                    """
                    The interpolation method needs self.theta to be in increasing order, so I
                    cannot use costheta directly.
                    """
                    theta[sel3d] = np.arccos((self._ss * yi[sel3d] - self._cs * zi[sel3d]) / dist3d[sel3d])
                    psca[sel3d,0] = np.interp(theta[sel3d], self.theta, self.s11[:,0])
                    psca[sel3d,1] = np.interp(theta[sel3d], self.theta, self.s11[:,1])
                    ppol[sel3d,0] = np.interp(theta[sel3d], self.theta, self.s12[:,0])
                    ppol[sel3d,1] = np.interp(theta[sel3d], self.theta, self.s12[:,1])
                """
                This is the number density for the scatterd light images
                """
                density[sel3d] = densr[sel3d] * densz[sel3d] * volume[sel3d] / (dist3d[sel3d]**2.)
                """
                Differentiate the north and south sides of the disk
                """
                sel = (self.azimuth <=0.)
                self.intensity[sel] += density[sel] * psca[sel,0]
                self.intensity[~sel] += density[~sel] * psca[~sel,1]
                self.polarized[sel] += density[sel] * ppol[sel,0]
                self.polarized[~sel] += density[~sel] * ppol[~sel,1]
                """
                Some debugging plots
                """
                #tmp = np.zeros(shape=(self._nx, self._nx))
                #tmp[sel] = density[sel] * psca[sel,0]
                #tmp[~sel] = density[~sel] * psca[~sel,1]

                #xlim = self._cx * self._pixscale
                #fig = plt.figure(figsize=(7,7 * 9./16.))
                #ax1 = fig.add_axes([0.0, 0.0, 1., 1.])
                #ax1.imshow(tmp, origin = 'lower', vmin = 0., vmax = 3.5e-7, cmap = 'inferno', extent=[xlim, -xlim, -xlim, xlim])
                #ax1.plot(0.,0., marker = '+', ms = 6 , color = 'w')
                #ax1.set_xlim(xlim, -xlim)
                #ax1.set_ylim(-xlim * 9./16., xlim * 9./16.)
                #ax1.axis('off')
                #plt.savefig('debug/image_'+format(i,'04d')+'.png', edgecolor='black', dpi=100, facecolor='black')
                ##if i==0:
                    ##plt.show()
                #plt.close()
        """
        For thermal emission, the flux is normalized to the total flux in the image,
        so that one can fit directly for the total flux in the image. Note that if not
        all the emission is contained in the image, you will miss some of the flux.
        """
        if self._thermal:
            self.intensity /= np.sum(self.intensity)

    """
    Plot the images
    """
    def plot(self, cmap = 'inferno'):
        """
        Method to make some plot
        """
        if self._thermal:
            fig = plt.figure(figsize=(7,7))
            ax1 = fig.add_axes([0.16, 0.14, 0.8, 0.79])
            im = ax1.imshow(self.intensity, origin = 'lower', extent = [self._xlim, -self._xlim, -self._xlim, self._xlim], cmap = cmap, vmin = np.percentile(self.intensity, 1.), vmax = np.percentile(self.intensity, 99.9))
            ax1.set_xlabel('$\Delta \\alpha$ [$^{\prime\prime}$]')
            ax1.set_ylabel('$\Delta \delta$ [$^{\prime\prime}$]')
            plt.show()
        else:
            fig = plt.figure(figsize=(13,6))
            ax1 = fig.add_axes([0.1, 0.12, 0.4, 0.8])
            im = ax1.imshow(self.intensity, origin = 'lower', extent = [self._xlim, -self._xlim, -self._xlim, self._xlim], cmap = cmap, vmin = np.percentile(self.intensity, 1.), vmax = np.percentile(self.intensity, 99.9))
            ax1.set_xlabel('$\Delta \\alpha$ [$^{\prime\prime}$]')
            ax1.set_ylabel('$\Delta \delta$ [$^{\prime\prime}$]')

            ax2 = fig.add_axes([0.55, 0.12, 0.4, 0.8])
            im = ax2.imshow(self.polarized, origin = 'lower', extent = [self._xlim, -self._xlim, -self._xlim, self._xlim], cmap = cmap, vmin = np.percentile(self.polarized, 1.), vmax = np.percentile(self.polarized, 99.9))
            ax2.set_xlabel('$\Delta \\alpha$ [$^{\prime\prime}$]')
            plt.show()

    """
    Check the parameters that are passed as kwargs
    """
    def _check_parameters(self, kwargs):
        """
        Check parameters that are being passed
        """
        if 'a' in kwargs:
            self._a = kwargs['a']
        if 'da' in kwargs:
            self._da = kwargs['da']
        if 'incl' in kwargs:
            self._incl = kwargs['incl'] * np.pi / 180.
        if 'pa' in kwargs:
            self._pa = -kwargs['pa'] * np.pi / 180.
        if 'pin' in kwargs:
            self._pin = kwargs['pin']
        if 'pmid' in kwargs:
            self._pmid = kwargs['pmid']
        if 'pout' in kwargs:
            self._pout = kwargs['pout']
        if 'gamma' in kwargs:
            self._gamma = kwargs['gamma']
        if 'gsca' in kwargs:
            self._gsca = kwargs['gsca']
        if 'gpol' in kwargs:
            self._gpol = kwargs['gpol']
        if 'e' in kwargs:
            self._e = kwargs['e']
        if 'omega' in kwargs:
            self._omega = kwargs['omega'] * np.pi / 180.
        if 'opang' in kwargs:
            self._opang = kwargs['opang']
        if 's11' in kwargs:
            self.s11 = kwargs['s11']
            self.theta = np.linspace(0., np.pi, num = np.shape(self.s11)[0])
            if 's12' not in kwargs: self.s12 = np.ones(shape=(np.shape(self.s11)[0],2))
        if 's12' in kwargs:
            self.s12 = kwargs['s12']
            self.theta = np.linspace(0., np.pi, num = np.shape(self.s12)[0])
            if 's11' not in kwargs: self.s11 = np.ones(shape=(np.shape(self.s12)[0],2))
        if 'wave' in kwargs:
            self._nu = CC / kwargs['wave']
        if ((self._thermal) and (self._nu is None)):
            self._error_msg('To compute thermal images, you need to pass a wavelength \'wave\' in units of microns.')
        if ((self._pmid is not None) and (self._da is None)):
            self._error_msg('If you need a third power-law you need to provide \'da\'')

    """
    Define some cos, sin, and tan
    """
    def _trigonometry(self):
        """
        Define some variables
        """
        if self._incl == 0.:
            self._incl = 0.1 * np.pi / 180.
        if self._incl == np.pi/2.:
            self._incl = 89.9 * np.pi / 180.
        self._cs, self._ss = np.cos(self._incl), np.sin(self._incl)
        self._st = self._cs / self._ss
        self._st2 = self._st**2.
        self._to = np.tan(self._opang)

    """
    Print some error message and quit
    """
    def _error_msg(self, message):
        """
        Print something and quit the program.
        """
        print(message)
        sys.exit()

if __name__ == '__main__':
    disk = Disk(nframe = 50, thermal = False, dpc= 71.)
    t0 = time.time()
    disk.compute_model(e = 0.2, incl = 70.1, pa = 110., a = 0.89, gsca = 0.4, gpol = 0.6, omega = 80., opang = 0.035, pin = 20.0, pout = -5.5, pmid = 0.5, da = 0.5)
    print('Took: ' + format(time.time()-t0, '0.2f') + ' seconds.')
    disk.plot()




"""
Small collection of robust statistical estimators based on functions from
Henry Freudenriech (Hughes STX) statistics library (called ROBLIB) that have
been incorporated into the AstroIDL User's Library.  Function included are:

  * medabsdev - median absolute deviation
  * biweightMean - biweighted mean estimator
  * mean - robust estimator of the mean of a data set
  * mode - robust estimate of the mode of a data set using the half-sample
    method
  * std - robust estimator of the standard deviation of a data set
  * checkfit - return the standard deviation and biweights for a fit in order 
    to determine its quality
  * linefit - outlier resistant fit of a line to data
  * polyfit - outlier resistant fit of a polynomial to data

For the fitting routines, the coefficients are returned in the same order as
np.polyfit, i.e., with the coefficient of the highest power listed first.

For additional information about the original IDL routines, see:
  http://idlastro.gsfc.nasa.gov/contents.html#C17
"""

from __future__ import division, print_function#, unicode_literals

import numpy as np
from numpy import median

import logging
_log = logging.getLogger('pynrc')

__version__ = '0.4'
__revision__ = '$Rev$'
__all__ = ['medabsdev','biweightMean', 'mean', 'mode', 'std', \
           'checkfit', 'linefit', 'polyfit', \
           '__version__', '__revision__', '__all__']


# Numerical precision
__epsilon = np.finfo(float).eps
__delta = 5.0e-7

def medabsdev(data, axis=None, keepdims=False, nan=True):
    """Median Absolute Deviation
    
    A "robust" version of standard deviation. Runtime is the 
    same as `astropy.stats.funcs.mad_std`.
    
    Parameters
    ----------
    data : ndarray
        The input data.
    axis : None or int or tuple of ints, optional
        Axis or axes along which the deviation is computed. The
        default is to compute the deviation of the flattened array.
        
        If this is a tuple of ints, a standard deviation is performed over
        multiple axes, instead of a single axis or all the axes as before.
        This is the equivalent of reshaping the input data and then taking
        the standard devation.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the original `arr`.
    nan : bool, optional
        Ignore NaNs? Default is True.
    """
    medfunc = np.nanmedian if nan else np.median
    meanfunc = np.nanmean if nan else np.mean

    if (axis is None) and (keepdims==False):
        data = data.ravel()
    
    # Scale factor to return result equivalent to standard deviation.
    sig_scale = 0.6744897501960817
    
    med = medfunc(data, axis=axis, keepdims=True)
    absdiff = np.abs(data - med)
    sigma = medfunc(absdiff, axis=axis, keepdims=True)  / sig_scale
    
    # Check if anything is near 0.0 (below machine precision)
    mask = sigma < __epsilon
    if np.any(mask):
        sigma[mask] = (meanfunc(absdiff, axis=axis, keepdims=True))[mask] / 0.8
    mask = sigma < __epsilon
    if np.any(mask):
        sigma[mask] = 0.0
        
        
    if len(sigma)==1:
        return sigma[0]
    elif not keepdims:
        return np.squeeze(sigma)
    else:
        return sigma



def biweightMean(inputData, axis=None, dtype=None, iterMax=25):
    """Biweight Mean
    
    Calculate the mean of a data set using bisquare weighting.  

    Based on the biweight_mean routine from the AstroIDL User's 
    Library.

    .. versionchanged:: 1.0.3
        Added the 'axis' and 'dtype' keywords to make this function more
        compatible with np.mean()
    """

    if axis is not None:
        fnc = lambda x: biweightMean(x, dtype=dtype)
        y0 = np.apply_along_axis(fnc, axis, inputData)
    else:
        y = inputData.ravel()
        if type(y).__name__ == "MaskedArray":
            y = y.compressed()
        if dtype is not None:
            y = y.astype(dtype)
        
        n = len(y)
        closeEnough = 0.03*np.sqrt(0.5/(n-1))
    
        diff = 1.0e30
        nIter = 0
    
        y0 = np.median(y)
        deviation = y - y0
        sigma = std(deviation)
    
        if sigma < __epsilon:
            diff = 0
        while diff > closeEnough:
            nIter = nIter + 1
            if nIter > iterMax:
                break
            uu = ((y-y0)/(6.0*sigma))**2.0
            uu = np.where(uu > 1.0, 1.0, uu)
            weights = (1.0-uu)**2.0
            weights /= weights.sum()
            y0 = (weights*y).sum()
            deviation = y - y0
            prevSigma = sigma
            sigma = std(deviation, Zero=True)
            if sigma > __epsilon:
                diff = np.abs(prevSigma - sigma) / prevSigma
            else:
                diff = 0.0
            
    return y0


def mean(inputData, Cut=3.0, axis=None, dtype=None, keepdims=False, 
    return_std=False, return_mask=False):
    """Robust Mean
    
    Robust estimator of the mean of a data set. Based on the `resistant_mean` 
    function from the AstroIDL User's Library. NaN values are excluded.

    This function trims away outliers using the median and the median 
    absolute deviation. An approximation formula is used to correct for
    the truncation caused by trimming away outliers.

    Parameters
    ==========
    inputData : ndarray
        The input data.

    Keyword Args
    ============
    Cut : float
        Sigma for rejection; default is 3.0.
    axis : None or int or tuple of ints, optional
        Axis or axes along which the deviation is computed. The
        default is to compute the deviation of the flattened array.
        
        If this is a tuple of ints, a standard deviation is performed over
        multiple axes, instead of a single axis or all the axes as before.
        This is the equivalent of reshaping the input data and then taking
        the standard devation.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the original `arr`.
    return_std : bool
        Also return the std dev calculated using only the "good" data?
    return_mask : bool
        If set to True, then return only boolean array of good (1) and 
        rejected (0) values.

    """

    inputData = np.array(inputData)
    
    if np.isnan(inputData).sum() > 0:
        medfunc = np.nanmedian
        meanfunc = np.nanmean
    else:
        medfunc = np.median
        meanfunc = np.mean

    if axis is None:
        data = inputData.ravel()
    else:
        data = inputData
        
    if type(data).__name__ == "MaskedArray":
        data = data.compressed()
    if dtype is not None:
        data = data.astype(dtype)

    # Scale factor to return result equivalent to standard deviation.
    sig_scale = 0.6744897501960817
        
    # Calculate the median absolute deviation
    data0 = medfunc(data, axis=axis, keepdims=True)
    absdiff = np.abs(data-data0)
    medAbsDev = medfunc(absdiff, axis=axis, keepdims=True) / sig_scale

    mask = medAbsDev < __epsilon
    if np.any(mask):
        medAbsDev[mask] = (meanfunc(absdiff, axis=axis, keepdims=True))[mask] / 0.8

    # First cut using the median absolute deviation
    cutOff = Cut*medAbsDev
    good = absdiff <= cutOff
    data_naned = data.copy()
    data_naned[~good] = np.nan
    dataMean = np.nanmean(data_naned, axis=axis, keepdims=True)
    dataSigma = np.nanstd(data_naned, axis=axis, keepdims=True)
    #dataSigma = np.sqrt( np.nansum((data_naned-dataMean)**2.0) / len(good) )

    # Calculate sigma
    if Cut > 1.0:
        sigmaCut = Cut
    else:
        sigmaCut = 1.0
    if sigmaCut <= 4.5:
        poly_sigcut = -0.15405 + 0.90723*sigmaCut - 0.23584*sigmaCut**2.0 + 0.020142*sigmaCut**3.0
        dataSigma = dataSigma / poly_sigcut

    cutOff = Cut*dataSigma
    good = absdiff <= cutOff

    if return_mask:
        return np.reshape(~np.isnan(data_naned), inputData.shape)

    data_naned = data.copy()
    data_naned[~good] = np.nan
    dataMean = np.nanmean(data_naned, axis=axis, keepdims=True)
    if return_std:
        dataSigma = np.nanstd(data_naned, axis=axis, keepdims=True)
    
    if len(dataMean)==1:
        if return_std:
            return dataMean[0], dataSigma[0]
        else:
            return dataMean[0]
    if not keepdims:
        if return_std:
            return np.squeeze(dataMean), np.squeeze(dataSigma)
        else:
            return np.squeeze(dataMean)
    else:
        if return_std:
            return dataMean, dataSigma
        else:
            return dataMean



def _mean_old(inputData, Cut=3.0, axis=None, dtype=None):
    """Robust mean
    
    Robust estimator of the mean of a data set.  Based on the 
    resistant_mean function from the AstroIDL User's Library.

    .. versionchanged:: 1.0.3
        Added the 'axis' and 'dtype' keywords to make this function more
        compatible with np.mean()
    """
           
    inputData = np.array(inputData)
    if axis is not None:
        fnc = lambda x: _mean_old(x, dtype=dtype)
        dataMean = np.apply_along_axis(fnc, axis, inputData)
    else:
        data = inputData.ravel()
        if type(data).__name__ == "MaskedArray":
            data = data.compressed()
        if dtype is not None:
            data = data.astype(dtype)
        
        data0 = np.median(data)
        absdiff = np.abs(data-data0)
        
        medAbsDev = np.median(absdiff) / 0.6745
        if medAbsDev < __epsilon:
            medAbsDev = absdiff.mean() / 0.8000
        
        cutOff = Cut*medAbsDev
        good = np.where( absdiff <= cutOff )
        good = good[0]
        dataMean = data[good].mean()
        dataSigma = np.sqrt( ((data[good]-dataMean)**2.0).sum() / len(good) )

        if Cut > 1.0:
            sigmaCut = Cut
        else:
            sigmaCut = 1.0
        if sigmaCut <= 4.5:
            dataSigma = dataSigma / (-0.15405 + 0.90723*sigmaCut - 0.23584*sigmaCut**2.0 + 0.020142*sigmaCut**3.0)
        
        cutOff = Cut*dataSigma
        good = np.where( absdiff <= cutOff )
        good = good[0]
        dataMean = data[good].mean()
        if len(good) > 3:
            dataSigma = np.sqrt( ((data[good]-dataMean)**2.0).sum() / len(good) )
        
        if Cut > 1.0:
            sigmaCut = Cut
        else:
            sigmaCut = 1.0
        if sigmaCut <= 4.5:
            dataSigma = dataSigma / (-0.15405 + 0.90723*sigmaCut - 0.23584*sigmaCut**2.0 + 0.020142*sigmaCut**3.0)
        
        dataSigma = dataSigma / np.sqrt(len(good)-1)
    
    return dataMean


def mode(inputData, axis=None, dtype=None):
    """
    Robust estimator of the mode of a data set using the half-sample mode.

    .. versionadded: 1.0.3
    """

    if axis is not None:
        fnc = lambda x: mode(x, dtype=dtype)
        dataMode = np.apply_along_axis(fnc, axis, inputData)
    else:
        # Create the function that we can use for the half-sample mode
        def _hsm(data):
            if data.size == 1:
                return data[0]
            elif data.size == 2:
                return data.mean()
            elif data.size == 3:
                i1 = data[1] - data[0]
                i2 = data[2] - data[1]
                if i1 < i2:
                    return data[:2].mean()
                elif i2 > i1:
                    return data[1:].mean()
                else:
                    return data[1]
            else:
                wMin = data[-1] - data[0]
                N = int(data.size/2 + data.size%2)
                for i in range(0, N):
                    w = data[i+N-1] - data[i] 
                    if w < wMin:
                        wMin = w
                        j = i
                return _hsm(data[j:j+N])
            
        data = inputData.ravel()
        if type(data).__name__ == "MaskedArray":
            data = data.compressed()
        if dtype is not None:
            data = data.astype(dtype)
        
        # The data need to be sorted for this to work
        data = np.sort(data)
    
        # Find the mode
        dataMode = _hsm(data)
    
    return dataMode

def std(inputData, Zero=False, axis=None, dtype=None, keepdims=False, return_mask=False):
    """Robust Sigma
    
    Based on the robust_sigma function from the AstroIDL User's Library.

    Calculate a resistant estimate of the dispersion of a distribution.
    
    Use the median absolute deviation as the initial estimate, then weight 
    points using Tukey's Biweight. See, for example, "Understanding Robust
    and Exploratory Data Analysis," by Hoaglin, Mosteller and Tukey, John
    Wiley & Sons, 1983, or equation 9 in Beers et al. (1990, AJ, 100, 32).

    Parameters
    ==========
    inputData : ndarray
        The input data.

    Keyword Args
    ============
    axis : None or int or tuple of ints, optional
        Axis or axes along which the deviation is computed. The
        default is to compute the deviation of the flattened array.
        
        If this is a tuple of ints, a standard deviation is performed over
        multiple axes, instead of a single axis or all the axes as before.
        This is the equivalent of reshaping the input data and then taking
        the standard devation.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the original `arr`.
    return_mask : bool
        If set to True, then only return boolean array of good (1) and 
        rejected (0) values.

	"""

    inputData = np.array(inputData)
    
    if np.isnan(inputData).sum() > 0:
        medfunc = np.nanmedian
        meanfunc = np.nanmean
    else:
        medfunc = np.median
        meanfunc = np.mean

    if axis is None:
        data = inputData.ravel()
    else:
        data = inputData

    if type(data).__name__ == "MaskedArray":
        data = data.compressed()
    if dtype is not None:
        data = data.astype(dtype)

    # Scale factor to return result equivalent to standard deviation.
    sig_scale = 0.6744897501960817

    # Calculate the median absolute deviation
    if Zero:
        data0 = 0.0
    else:
        data0 = medfunc(data, axis=axis, keepdims=True)
    absdiff = np.abs(data-data0)
    medAbsDev = medfunc(absdiff, axis=axis, keepdims=True) / sig_scale
    mask = medAbsDev < __epsilon
    if np.any(mask):
        medAbsDev[mask] = (meanfunc(absdiff, axis=axis, keepdims=True))[mask] / 0.8
        
    # These will be set to 0 later
    mask0 = medAbsDev < __epsilon
        
    u = (data-data0) / (6.0 * medAbsDev)
    u2 = u**2.0
    good = u2 <= 1.0

    if return_mask:
        return good & ~np.isnan(data)
    
    # These values will be set to NaN later
    # if fewer than 3 good points to calculate stdev
    ngood = good.sum(axis=axis, keepdims=True)
    mask_nan = ngood < 2
    if mask_nan.sum() > 0:
        print("WARNING: NaN's will be present due to weird distributions")
    
    # Set bad points to NaNs
    u2[~good] = np.nan
    
    numerator = np.nansum( (data - data0)**2 * (1.0 - u2)**4.0, axis=axis, keepdims=True)
    nElements = len(data) if axis is None else data.shape[axis]
    denominator = np.nansum( (1.0 - u2) * (1.0 - 5.0*u2), axis=axis, keepdims=True)
    sigma = np.sqrt( nElements*numerator / (denominator*(denominator-1.0)) )
    
    sigma[mask0] = 0
    sigma[mask_nan] = np.nan

    if len(sigma)==1:
        return sigma[0]
    elif not keepdims:
        return np.squeeze(sigma)
    else:
        return sigma


def _std_old(inputData, Zero=False, axis=None, dtype=None):
    """
    Robust estimator of the standard deviation of a data set.  

    Based on the robust_sigma function from the AstroIDL User's Library.

    .. versionchanged:: 1.0.3
        Added the 'axis' and 'dtype' keywords to make this function more
        compatible with np.std()
    """

    if axis is not None:
        fnc = lambda x: _std_old(x, dtype=dtype)
        sigma = np.apply_along_axis(fnc, axis, inputData)
    else:
        data = inputData.ravel()
        if type(data).__name__ == "MaskedArray":
            data = data.compressed()
        if dtype is not None:
            data = data.astype(dtype)
        
        if Zero:
            data0 = 0.0
        else:
            data0 = np.median(data)
        medAbsDev = np.median(np.abs(data-data0)) / 0.6744897501960817
        if medAbsDev < __epsilon:
            medAbsDev = np.mean(np.abs(data-data0)) / 0.8000
        if medAbsDev < __epsilon:
            sigma = 0.0
            return sigma
        
        u = (data-data0) / 6.0 / medAbsDev
        u2 = u**2.0
        good = np.where( u2 <= 1.0 )
        good = good[0]
        if len(good) < 3:
            print("WARNING:  Distribution is too strange to compute standard deviation")
            sigma = -1.0
            return sigma
        
        numerator = ((data[good]-data0)**2.0 * (1.0-u2[good])**2.0).sum()
        nElements = (data.ravel()).shape[0]
        denominator = ((1.0-u2[good])*(1.0-5.0*u2[good])).sum()
        sigma = nElements*numerator / (denominator*(denominator-1.0))
        if sigma > 0:
            sigma = np.sqrt(sigma)
        else:
            sigma = 0.0
        
    return sigma


def checkfit(inputData, inputFit, epsilon, delta, BisquareLimit=6.0):
    """
    Determine the quality of a fit and biweights.  Returns a tuple
    with elements:
    
      0. Robust standard deviation analog
      1. Fractional median absolute deviation of the residuals
      2. Number of input points given non-zero weight in the calculation
      3. Bisquare weights of the input points
      4. Residual values scaled by sigma

    This function is based on the rob_checkfit routine from the AstroIDL 
    User's Library.
    """

    data = inputData.ravel()
    fit = inputFit.ravel()
    if type(data).__name__ == "MaskedArray":
        data = data.compressed()
    if type(fit).__name__ == "MaskedArray":
        fit = fit.compressed()

    deviation = data - fit
    sigma = std(deviation, Zero=True)
    if sigma < epsilon:
        return (sigma, 0.0, 0, 0.0, 0.0)

    toUse = (np.where( np.abs(fit) > epsilon ))[0]
    if len(toUse) < 3:
        fracDev = 0.0
    else:
        fracDev = np.median(np.abs(deviation[toUse]/fit[toUse]))
    if fracDev < delta:
        return (sigma, fracDev, 0, 0.0, 0.0)
    
    biweights = np.abs(deviation)/(BisquareLimit*sigma)
    toUse = (np.where(biweights > 1))[0]
    if len(toUse) > 0:
        biweights[toUse] = 1.0
    nGood = len(data) - len(toUse)

    scaledResids = (1.0 - biweights**2.0)
    scaledResids = scaledResids / scaledResids.sum()

    return (sigma, fracDev, nGood, biweights, scaledResids)


def linefit(inputX, inputY, iterMax=25, Bisector=False, BisquareLimit=6.0, CloseFactor=0.03):
    """
    Outlier resistance two-variable linear regression function.

    Based on the robust_linefit routine in the AstroIDL User's Library.
    """

    xIn = inputX.ravel()
    yIn = inputY.ravel()
    if type(yIn).__name__ == "MaskedArray":
        xIn = xIn.compress(np.logical_not(yIn.mask))
        yIn = yIn.compressed()
    n = len(xIn)

    np.logical_not(yIn.mask)

    x0 = xIn.sum() / n
    y0 = yIn.sum() / n
    x = xIn - x0
    y = yIn - y0

    cc = np.zeros(2)
    ss = np.zeros(2)
    sigma = 0.0
    yFit = yIn
    badFit = 0
    nGood = n

    lsq = 0.0
    yp = y
    if n > 5:
        s = np.argsort(x)
        u = x[s]
        v = y[s]
        nHalf = n//2 -1
        x1 = np.median(u[0:nHalf])
        x2 = np.median(u[nHalf:])
        y1 = np.median(v[0:nHalf])
        y2 = np.median(v[nHalf:])
        if np.abs(x2-x1) < __epsilon:
            x1 = u[0]
            x2 = u[-1]
            y1 = v[0]
            y2 = v[-1]
        cc[1] = (y2-y1)/(x2-x1)
        cc[0] = y1 - cc[1]*x1
        yFit = cc[0] + cc[1]*x
        sigma, fracDev, nGood, biweights, scaledResids = checkfit(yp, yFit, __epsilon, __delta)
        if nGood < 2:
            lsq = 1.0
        
    if lsq == 1 or n < 6:
        sx = x.sum()
        sy = y.sum()
        sxy = (x*y).sum()
        sxx = (x*x).sum()
        d = sxx - sx*sx
        if np.abs(d) < __epsilon:
            return (0.0, 0.0)
        ySlope = (sxy - sx*sy) / d
        yYInt = (sxx*sy - sx*sxy) / d
    
        if Bisector:
            syy = (y*y).sum()
            d = syy - sy*sy
            if np.abs(d) < __epsilon:
                return (0.0, 0.0)
            tSlope = (sxy - sy*sx) / d
            tYInt = (syy*sx - sy*sxy) / d
            if np.abs(tSlope) < __epsilon:
                return (0.0, 0.0)
            xSlope = 1.0/tSlope
            xYInt = -tYInt / tSlope
            if ySlope > xSlope:
                a1 = yYInt
                b1 = ySlope
                r1 = np.sqrt(1.0+ySlope**2.0)
                a2 = xYInt
                b2 = xSlope
                r2 = np.sqrt(1.0+xSlope**2.0)
            else:
                a2 = yYInt
                b2 = ySlope
                r2 = np.sqrt(1.0+ySlope**2.0)
                a1 = xYInt
                b1 = xSlope
                r1 = np.sqrt(1.0+xSlope**2.0)
            yInt = (r1*a2 + r2*a1) / (r1 + r2)
            slope = (r1*b2 + r2*b1) / (r1 + r2)
            r = np.sqrt(1.0+slope**2.0)
            if yInt > 0:
                r = -r
            u1 = slope / r
            u2 = -1.0/r
            u3 = yInt / r
            yp = u1*x + u2*y + u3
            yFit = y*0.0
            ss = yp
        else:
            slope = ySlope
            yInt = yYInt
            yFit = yInt + slope*x
        cc[0] = yInt
        cc[1] = slope
        sigma, fracDev, nGood, biweights, scaledResids = checkfit(yp, yFit, __epsilon, __delta)
    
    if nGood < 2:
        cc[0] = cc[0] + y0 - cc[1]*x0
        return cc[::-1]
    
    sigma1 = (100.0*sigma)
    closeEnough = CloseFactor * np.sqrt(0.5/(n-1))
    if closeEnough < __delta:
        closeEnough = __delta
    diff = 1.0e20
    nIter = 0
    while diff > closeEnough:
        nIter = nIter + 1
        if nIter > iterMax:
            break
        sigma2 = sigma1
        sigma1 = sigma
        sx = (biweights*x).sum()
        sy = (biweights*y).sum()
        sxy = (biweights*x*y).sum()
        sxx = (biweights*x*x).sum()
        d = sxx - sx*sx
        if np.abs(d) < __epsilon:
            return (0.0, 0.0)
        ySlope = (sxy - sx*sy) / d
        yYInt = (sxx*sy - sx*sxy) / d
        slope = ySlope
        yInt = yYInt
    
        if Bisector:
            syy = (biweights*y*y).sum()
            d = syy - sy*sy
            if np.abs(d) < __epsilon:
                return (0.0, 0.0)
            tSlope = (sxy - sy*sx) / d
            tYInt = (syy*sx - sy*sxy) / d
            if np.abs(tSlope) < __epsilon:
                return (0.0, 0.0)
            xSlope = 1.0/tSlope
            xYInt = -tYInt / tSlope
            if ySlope > xSlope:
                a1 = yYInt
                b1 = ySlope
                r1 = np.sqrt(1.0+ySlope**2.0)
                a2 = xYInt
                b2 = xSlope
                r2 = np.sqrt(1.0+xSlope**2.0)
            else:
                a2 = yYInt
                b2 = ySlope
                r2 = np.sqrt(1.0+ySlope**2.0)
                a1 = xYInt
                b1 = xSlope
                r1 = np.sqrt(1.0+xSlope**2.0)
            yInt = (r1*a2 + r2*a1) / (r1 + r2)
            slope = (r1*b2 + r2*b1) / (r1 + r2)
            r = np.sqrt(1.0+slope**2.0)
            if yInt > 0:
                r = -r
            u1 = slope / r
            u2 = -1.0/r
            u3 = yInt / r
            yp = u1*x + u2*y + u3
            yFit = y*0.0
            ss = yp
        else:
            yFit = yInt + slope*x
        cc[0] = yInt
        cc[1] = slope
        sigma, fracDev, nGood, biweights, scaledResids = checkfit(yp, yFit, __epsilon, __delta)
    
        if nGood < 2:
            badFit = 1
            break
        diff1 = np.abs(sigma1 - sigma)/sigma
        diff2 = np.abs(sigma2 - sigma)/sigma
        if diff1 < diff2:
            diff = diff1
        else:
            diff = diff2
        
    cc[0] = cc[0] + y0 - cc[1]*x0
    return cc[::-1]


def polyfit(inputX, inputY, order, iterMax=25):
    """
    Outlier resistance two-variable polynomial function fitter.

    Based on the robust_poly_fit routine in the AstroIDL User's 
    Library.

    Unlike robust_poly_fit, two different polynomial fitters are used
    because np.polyfit does not support non-uniform weighting of the
    data.  For the weighted fitting, the SciPy Orthogonal Distance
    Regression module (scipy.odr) is used.
    """

    from scipy import odr

    def polyFunc(B, x, order=order):
        out = x*0.0
        for i in range(order+1):
            out = out + B[i]*x**i

    model = odr.Model(polyFunc)

    x = inputX.ravel()
    y = inputY.ravel()
    if type(y).__name__ == "MaskedArray":
        x = x.compress(np.logical_not(y.mask))
        y = y.compressed()
    n = len(x)

    x0 = x.sum() / n
    y0 = y.sum() / n
    u = x
    v = y

    nSeg = int(order + 2)
    if (nSeg/2)*2 == nSeg:
        nSeg = nSeg + 1
    minPts = nSeg*3
    if n < 1000:
        lsqFit = 1
        cc = np.polyfit(u, v, order)
        yFit = np.polyval(cc, u)
    else:
        lsqfit = 0
        q = np.argsort(u)
        u = u[q]
        v = v[q]
        nPerSeg = np.zeros(nSeg, dtype='int') + n//nSeg
        nLeft = n - nPerSeg[0]*nSeg
        nPerSeg[nSeg//2] = nPerSeg[nSeg//2] + nLeft
        r = np.zeros(nSeg)
        s = np.zeros(nSeg)
        print(nPerSeg)
        r[0] = np.median(u[0:nPerSeg[0]])
        s[0] = np.median(v[0:nPerSeg[0]])
        i2 = nPerSeg[0]-1
        for i in range(1,nSeg):
            i1 = i2
            i2 = i1 + nPerSeg[i]
            r[i] = np.median(u[i1:i2])
            s[i] = np.median(v[i1:i2])
        cc = np.polyfit(r, s, order)
        yFit = np.polyval(cc, u)
    
    sigma, fracDev, nGood, biweights, scaledResids = checkfit(v, yFit, __epsilon, __delta)
    if nGood == 0:
        return cc
    if nGood < minPts:
        if lsqFit == 0:
            cc = np.polyfit(u, v, order)
            yFit = np.polyval(cc, u)
            sigma, fracDev, nGood, biweights, scaledResids = checkfit(yp, yFit, __epsilon, __delta)
            if nGood == 0:
                return __processPoly(x0, y0, order, cc)
            nGood = n - nGood
        if nGood < minPts:
            return 0
        
    closeEnough = 0.03*np.sqrt(0.5/(n-1))
    if closeEnough < __delta:
        closeEnough = __delta
    diff = 1.0e10
    sigma1 = 100.0*sigma
    nIter = 0
    while diff > closeEnough:
        nIter = nIter + 1
        if nIter > iterMax:
            break
        sigma2 = sigma1
        sigma1 = sigma
        g = (np.where(biweights > 0))[0]
        if len(g) < len(biweights):
            u = u[g]
            v = v[g]
            biweights = biweights[g]
        try:
            ## Try the fancy method...
            data = odr.RealData(u, v, sy=1.0/biweights)
            fit = odr.ODR(data, model, beta0=cc[::-1])
            out = fit.run()
            cc = out.beta[::-1]
        except:
            ## And then give up when it doesn't work
            cc = np.polyfit(u, v, order)
        yFit = np.polyval(cc, u)
        sigma, fracDev, nGood, biweights, scaledResids = checkfit(v, yFit, __epsilon, __delta)
        if nGood < minPts:
            return cc
        diff1 = np.abs(sigma1 - sigma)/sigma
        diff2 = np.abs(sigma2 - sigma)/sigma
        if diff1 < diff2:
            diff = diff1
        else:
            diff = diff2
    return cc
